<?php
/*
 Main Canada Post Webserivce Class
 woocommerce_cpwebservice.php

Copyright (c) 2015 Jamez Picard

*/
class woocommerce_cpwebservice extends WC_Shipping_Method
{	
	
	/**
	 * __construct function.
	 *
	 * @access public
	 * @return woocommerce_cpwebservice
	 */
	function __construct() {
		$this->id 			= 'woocommerce_cpwebservice';
		$this->method_title 	= __('Canada Post', 'woocommerce-canadapost-webservice');
		$this->method_description 		= __( 'Canada Post Shipping Method.', 'woocommerce' );
		$this->init();
	}

	/**
	 * init function.
	 *
	 * @access public
	 * @return void
	 */
	function init() {
		$default_options = (object) array('enabled'=>'no', 'title'=>'Canada Post', 'api_user'=>'', 'api_key'=>'','account'=>'','contractid'=>'','source_postalcode'=>'','mode'=>'live', 'display_errors'=>false, 'prefer_service'=>false,
						   'delivery'=>'', 'delivery_guarantee'=>false, 'margin'=>'', 'margin_value'=>'', 'packageweight'=>floatval('0.02'), 'boxes_enable'=> false, 'boxes_switch'=>true, 'lettermail_enable'=> false, 'rules_enable'=>false, 'volumetric_weight'=>true, 'product_shipping_options' => true, 'weight_only_enabled'=> true,
						   'shipping_tracking'=> true, 'email_tracking'=> true, 'log_enable'=>false,'lettermail_limits'=>false,'lettermail_maxlength'=>'','lettermail_maxwidth'=>'','lettermail_maxheight'=>'','lettermail_override_weight'=>false,'lettermail_packageweight'=>'', 'lettermail_exclude_tax'=> false, 'tracking_icons'=> true, 'display_required_notice'=>true);
		$this->options		  = get_option('woocommerce_cpwebservice', $default_options);
		$this->options		  =	(object) array_merge((array) $default_options, (array) $this->options); // ensure all keys exist, as defined in default_options.
		$this->enabled		  = $this->options->enabled;
		$this->title 		  = $this->options->title;
		$this->type			  = 'order';
		$this->boxes		  = get_option('woocommerce_cpwebservice_boxes');
		$this->services		  = get_option('woocommerce_cpwebservice_services', array());
		$this->lettermail	  = get_option('woocommerce_cpwebservice_lettermail', array());
		$this->rules		  = get_option('woocommerce_cpwebservice_rules', array());
		$this->log 			  = (object) array('cart'=>array(),'params'=>array(),'request'=>array('http'=>'','service'=>''),'rates'=>array());

		// Defined Services
		$this->init_available_services();

		// Actions
		add_action('woocommerce_update_options_shipping_' . $this->id, array(&$this, 'process_admin_options'));
	}


		
		
	/*
	 * Defined Services
	*/
	function init_available_services() {
		$this->available_services = array(
				'DOM.RP'=>__('Regular Parcel', 'woocommerce-canadapost-webservice'),
				'DOM.EP'=>__('Expedited Parcel', 'woocommerce-canadapost-webservice'),
				'DOM.XP'=>__('Xpresspost', 'woocommerce-canadapost-webservice'),
				'DOM.XP.CERT'=>__('Xpresspost Certified', 'woocommerce-canadapost-webservice'),
				'DOM.PC'=>__('Priority', 'woocommerce-canadapost-webservice'),
				'DOM.LIB'=>__('Library Books', 'woocommerce-canadapost-webservice'),
				'USA.SP.AIR'=>__('Small Packet USA Air', 'woocommerce-canadapost-webservice'),
				'USA.TP'=>__('Tracked Packet USA', 'woocommerce-canadapost-webservice'),
				'USA.EP'=>__('Expedited Parcel USA', 'woocommerce-canadapost-webservice'),
				'USA.XP'=>__('Xpresspost USA', 'woocommerce-canadapost-webservice'),
				'USA.PW.ENV'=>__('Priority Worldwide Envelope USA', 'woocommerce-canadapost-webservice'),
				'USA.PW.PAK'=>__('Priority Worldwide Pak USA', 'woocommerce-canadapost-webservice'),
				'USA.PW.PARCEL'=>__('Priority Worldwide Parcel USA', 'woocommerce-canadapost-webservice'),
				'INT.SP.SURF'=>__('Small Packet International Surface', 'woocommerce-canadapost-webservice'),
				'INT.SP.AIR'=>__('Small Packet International Air', 'woocommerce-canadapost-webservice'),
				'INT.IP.AIR'=>__('International Parcel Air', 'woocommerce-canadapost-webservice'),
				'INT.IP.SURF'=>__('International Parcel Surface', 'woocommerce-canadapost-webservice'),
				'INT.TP'=>__('Tracked Packet International', 'woocommerce-canadapost-webservice'),
				'INT.XP'=>__('Xpresspost International', 'woocommerce-canadapost-webservice'),
				'INT.PW.ENV'=>__('Priority Worldwide Envelope International', 'woocommerce-canadapost-webservice'),
				'INT.PW.PAK'=>__('Priority Worldwide Pak International', 'woocommerce-canadapost-webservice'),
				'INT.PW.PARCEL'=>__('Priority Worldwide parcel International', 'woocommerce-canadapost-webservice')
		);
	}
		

	function get_destination_from_service($service_code){
		if (!empty($service_code) && strlen($service_code) >= 3) {
			switch(substr($service_code,0,3)) {
				case 'DOM': return __('Canada', 'woocommerce-canadapost-webservice');
				case 'USA': return __('USA', 'woocommerce-canadapost-webservice');
				case 'INT': return __('International', 'woocommerce-canadapost-webservice');
			}
		}
		return '';
	}
		
	/*
	 * Main Canada Post Lookup Rates function
	*/
	function calculate_shipping( $package = array() ) {
			
		global $woocommerce;
		
		// Need to calculate total package weight.

		// Get total volumetric weight.
		$total_quantity = 0;
		$total_weight = 0;
		$max = array('length'=>0, 'width'=>0, 'height'=>0);
		$dimension_unit = get_option( 'woocommerce_dimension_unit' );
		$weight_unit = get_option( 'woocommerce_weight_unit' );
		$item_weight = 0;
		$length = $width = $height = 0;
		$cubic = 0;
		$products = array();
	
		foreach ( $package['contents'] as $item_id => $values ) {
			if ( $values['quantity'] > 0 && $values['data']->needs_shipping() && $values['data']->has_weight() ) {
				$total_quantity += $values['quantity'];
				$item_weight = $this->convertWeight($values['data']->get_weight(), $weight_unit);
				$total_weight +=  $item_weight * $values['quantity'];
				$length = 0; $width = 0; $height = 0;
				if ( $values['data']->has_dimensions() ) {
					$dimensions = explode(' x ',str_replace($dimension_unit,'',$values['data']->get_dimensions()));
					if (count($dimensions) >= 3) {
						// Get cubic size.
						$length = $this->convertSize($dimensions[0], $dimension_unit);
						$width = $this->convertSize( $dimensions[1], $dimension_unit);
						$height = $this->convertSize( $dimensions[2], $dimension_unit);
						//  Rotate so that height the smallest dimension. Makes for consistent and efficient packing.
						$this->rotate_package($length, $width, $height);
					}
				
				} // Allow products with weight-only (no dimensions) to still be calculated. 
				if ($length == 0 && $width==0 && $height == 0 && $item_weight > 0 && $this->options->weight_only_enabled) {
					// Weight-only so use reverse-volumetric weight. 
					// volumetric weight = cubic / 6000 : 0; //Canada Post: (L cm x W cm x H cm)/6000
					$volumetric_cubic = $item_weight * 6000; 
					// cube Root volume instead of Boxes. (item is assumed already packaged to ship)
					$dimension = (float)pow($volumetric_cubic, 1.0/3.0);
					// A cube is the best estimate we can make.
					$length = $width = $height = number_format($dimension,2);
				}
				
				// Max dimensions
				if ($length > $max['length']) {  $max['length'] = $length; }
				if ($width > $max['width']) {  $max['width'] = $width; }
				if ($height > $max['height']) {  $max['height'] = $height; }
				
				// Add to Products Array
				if ($length > 0 && $width > 0 && $height > 0){
					$product_id = ($values['data']->is_type('variation') ? $values['data']->get_variation_id(): $values['data']->id);
					// Lookup custom product options
					$product_shipping_prepacked = get_post_meta( $product_id, '_cpwebservice_product_shipping', true );
					for($j=0;$j<intval($values['quantity']);$j++){
						$products[] = array('length'=>$length, 'width'=>$width, 'height'=>$height, 'item_id'=> $product_id,  'weight'=> $item_weight, 'cubic'=>($length * $width * $height), 'prepacked'=>$product_shipping_prepacked);
					}
				}
					
					// Cubic size
					$cubic +=  $length * $width * $height * $values['quantity'];

				// Cart Logging
				if ($this->options->log_enable){
					$this->log->cart[] = array('id'=>$values['data']->id, 'item'=>$values['data']->get_title(),'quantity'=>$values['quantity'], 'weight'=>$item_weight * $values['quantity'],
							'length'=>$length, 'width'=>$width, 'height'=>$height, 'cubic'=>($length * $width * $height * $values['quantity']));
				}
			}
		}
		
		// Box packing!
		$pack = new woocommerce_cpwebservice_pack();
		
		// Max box size (container in cm and kg) as defined on Canada Post's documentation.
		// http://www.canadapost.ca/cpo/mc/assets/pdf/business/parcelserviceguide_en.pdf (page 43)
		// Girth: length + (height x 2) + (width x 2)
		// No one dimension may exceed 2 m (200 cm)
		// Max. length + girth = 3 m (300 cm)
		$max_cp_box = array('length'=>200 , 'width'=> 200, 'height'=>200, 'girth'=> 300, 'weight'=> 30);
		
		// Max number rate lookups (To avoid webservice usage issues)
		$max_cp_lookups = 10;
		
		// Packages
		$containers = array();
		
		// Envelope weight with bill/notes/advertising inserts: ex. 20g
		$max_cp_box['weight'] -= (!empty($this->options->packageweight) ? floatval($this->options->packageweight) : 0);

		// Add Pre-packed products
		$products_tmp = array();
		foreach($products as $i => $p ){
			if (isset($p['prepacked']) && $p['prepacked']=='yes') {
				$containers[] = $this->to_container($p);
			} else {
				$products_tmp[] = $p;
			}
			if ($p['weight'] > $max_cp_box['weight']){
				// Cart has a product that is too heavy to pack or even mail on its own. Cancel lookups.
				$products = array();
			}
		}
		
		// Exclude products that are pre-packed
		if (count($containers) > 0){
			$products = $products_tmp;
		}
		
		// Sort products for consistant packing.
		usort($products, array(&$this, 'sort_products'));
		
		// Sort boxes if defined
		$max_defined_box = null;
		$pack_with_boxes = ($this->options->boxes_enable && is_array($this->boxes));
		$box_switch = false; // allows the situation when products don't fit in a defined box to switch to optimistic packing.
		if($this->options->boxes_enable && is_array($this->boxes)){
			// rotate boxes so smallest dimension is height.
			$this->rotate_boxes($this->boxes);
			// usort(boxes)
			usort($this->boxes, array(&$this, 'sort_boxes'));
			// used for packing
			$max_defined_box =  array('length'=>$max_cp_box['length'] , 'width'=> $max_cp_box['width'], 'height'=>$max_cp_box['height'], 'girth'=> $max_cp_box['girth'], 'weight'=> $max_cp_box['weight']);
		}
		
		// Loop over products to pack. 
		while (!empty($products) && is_array($products)){
		
			// Optimistic pack - no box containers defined.
			if (!$pack_with_boxes || $box_switch){
				$products_topack = count($products);
				// $box_switch = false; // switch back to box mode if that's why it's here.  
				// Pack and Check Container
				$pack->pack($products, null, $max_cp_box);
				$packed_boxes = $pack->get_packed_boxes();
				if (is_array($packed_boxes) && count($packed_boxes) > 0){
					$container = $pack->get_container_dimensions(); // return array('length' => 0,'width' => 0,'height' => 0);
					$container['cubic'] = $pack->get_container_volume(); // number
					
					
					$container['products'] = $packed_boxes;
					
					// Add to containers array.
					$containers[] = $container;
					
					// Loop to pack remaining products
					$products = $pack->get_remaining_boxes();
				}// endif
				
				if ($products_topack == count($products) && $products_topack > 0){
					// No products were packed this loop. (probably too big or too heavy, hitting the max)  Have to finish this while loop.
					// Break from loop and... ship products individually. Add remaining products as containers themselves
					$p_containers = array_map(array(&$this,'to_container'), $products);
					// Add remaining products as containers themselves.
					foreach($p_containers as $c){
						$containers[] = $c;
					}
					break;
				}
				
			} elseif($pack_with_boxes) {
				$products_topack = count($products);
				// Loop [get remaining products] and start with smallest box and go up to biggest box
				// Boxes have been sorted from smallest to biggest already.
				$last_box = count($this->boxes);
				$i = 0;
				foreach ($this->boxes as $box){
					$i++;
					if (!empty($box) && is_array($box)){
						if ($i==$last_box){ $max_defined_box['height'] = $box['height'];  } // last/largest box constrain height.
						// Pack and Check Container
						$pack->pack($products, $box, $max_defined_box);
						$packed_boxes = $pack->get_packed_boxes();
						if (is_array($packed_boxes) && count($packed_boxes) > 0){
							// Great! Products were packed in this box.
							$container = $pack->get_container_dimensions(); // return array('length' => 0,'width' => 0,'height' => 0);
							if (floatval($container['height']) > floatval($box['height'])) {
								// Because 'height' is dynamically calculated so we have made sure it actually fits inside this box)
								// If the dynamic height is greater than the box, we need to skip this box.
								continue;
							}
							$container['cubic'] = $pack->get_container_volume(); // number
							$container['box_weight'] = $box['weight'];
							$container['box_margin'] = $box['margin'];
							$container['box_name'] 	 = $box['name'];
							
							$container['products'] = $packed_boxes;
							
							// If the box has weight, add it to the shipping weight.
							if (floatval($container['box_weight']) > 0){
								$container['weight'] += floatval($container['box_weight']);
							}
							
							// Container height is dynamic in the box-packing algorithm.  Set to the defined box height, because that's the true package size.
							$container['height'] = $box['height'];
							// Recalc cubic because of height change.
							$container['cubic'] = $container['length'] * $container['width'] * $container['height'];
							
							// Add to containers array.
							$containers[] = $container;
							
							// Loop to pack remaining products
							$products = $pack->get_remaining_boxes();
						} // end if
						// If none remaining, break from foreach.
						if (!is_array($products) || count($products) == 0){
							break;
						}
					} // endif
					// else loop and try next smallest box.
					
				} // end foreach
				
				if ($products_topack == count($products) && $products_topack > 0){
					// No products were packed this loop. (probably too big or too heavy, hitting the max)  Have to finish this while loop.
					// Switch to non-box method? (TODO Configurable in options)
					if ($this->options->boxes_switch){
						$box_switch = true;
						continue;
					} else {
						// Alternatively, break from loop and... ship products individually. Add remaining products as containers themselves
						array_push($containers, array_map(array(&$this,'to_container'), $products));
						break;
					}
				} // endif
			} // endif
		
		} // end while
		
		// Done Box Packing!
		$legacy_volumetric_packing = false;
		if ($legacy_volumetric_packing){
		
			// Find which box the items will fit in (by cubic + packaging factor).
			$box_fits = null; // to fit.
	
			if ($this->options->boxes_enable && is_array($this->boxes)){
				foreach ($this->boxes as $box){
					$box_cubed = $box['cubed'];
					if ($cubic < $box_cubed && $this->boxFits($box,$max)){ // volume fitting and check for max dimension fitting.
						// It Fits!
						if (empty($box_fits)) {
							$box_fits = $box;
							// Use if smaller than previously iterated box that fits.
						} elseif (is_array($box_fits) && $box_cubed < $box_fits['cubed']) {
							$box_fits = $box;
						}
					}
				}
			}
			
			if (empty($box_fits)) { // If box was not found or boxes are not enabled
				if (count($products)>1){
					// Method: Cube Root volume instead of Boxes. (item is assumed already packaged to ship)
					$dimension = (float)pow($cubic, 1.0/3.0);
					// use max dimensions to ensure an item like 1x1x20 is estimated with enough length.
					$box_fits = array('cubed'=>$cubic, 'length'=>($dimension < $max['length'] ? $max['length'] : $dimension), 'width'=>($dimension < $max['width'] ? $max['width'] : $dimension), 'height'=>($dimension < $max['height'] ? $max['height'] : $dimension));
				} elseif(count($products)==1) {
					// only one product
					$box_fits = array('cubed'=>$cubic, 'length'=> $products[0]['length'], 'width'=> $products[0]['width'], 'height'=> $products[0]['height']);
				}
			}
			
			// Calculate product weight.
			$packed_boxes_weight = floatval(0);
			foreach($products as $p ){
				$packed_boxes_weight += isset($p['weight']) ? floatval($p['weight']) : 0;
			}
			$box_fits['weight'] = $packed_boxes_weight + (isset($box_fits['weight']) ? floatval($box_fits['weight']) : 0); // Also add box weight if it exists. 
			
			$container = array($box_fits);
		
		} // end if

		// Destination information (Need this to calculate rate)
		$country = $package['destination']['country']; // 2 char country code
		$state = $package['destination']['state']; // 2 char prov/state code
		$postal = $package['destination']['postcode']; // postalcode/zip code as entered by user.

		// Get a rate to ship the package.
		if ($country != '' && is_array($containers)) {
			
			// Loop $containers to get_rates.
			$rates = array();
			$rates_combined = array();
			$cp_lookups = 0;
			// Ensure service codes are in _all_ rate objects
			$distinct_service_codes = array();
			foreach($containers as $i=>$shipping_package){
				$total_weight = ($shipping_package['weight'] > 0) ? $shipping_package['weight'] : 0;
				if ($this->options->volumetric_weight){
					$volumetric_weight = $shipping_package['cubic'] > 0 ? $shipping_package['cubic'] / 6000 : 0; //Canada Post: (L cm x W cm x H cm)/6000
					// Use the largest value of total weight or volumetric/dimensional weight
					$shipping_weight = ($total_weight <= $volumetric_weight && $volumetric_weight <= $max_cp_box['weight']) ? $volumetric_weight : $total_weight;
				} else {
					$shipping_weight = $total_weight;
				}
				// Envelope weight with bill/notes/advertising inserts: ex. 20g
				$shipping_weight += (!empty($this->options->packageweight) ? floatval($this->options->packageweight) : 0);
				// update weight
				$containers[$i]['actual_weight'] = $shipping_package['weight'];
				$containers[$i]['weight'] = $shipping_weight;
				$limit_services = null;
				if (isset($this->options->rules_enable) && $this->options->rules_enable){
					$limit_services = $this->get_limited_services($shipping_package['products']);
				}
				
				$shipping_weight = round($shipping_weight,2); // 2 decimal places.
				$length = round($shipping_package['length'], 2);
				$width = round($shipping_package['width'], 2);
				$height = round($shipping_package['height'], 2);
				
				// Debug
				if ($this->options->log_enable){
					$this->log->params[] = array('country'=>$country, 'state'=>$state, 'postal'=>$postal, 'shipping_weight'=>$shipping_weight, 'total_weight'=>$total_weight, 'length'=>$length, 'width'=>$width, 'height'=>$height, 'box_name'=>(isset($shipping_package['box_name']) ? $shipping_package['box_name'] : ''));
				}
				
				// In subsequent iterations, use Rates from a previous package/container that is identical to current container. Else do a rates lookup.
				$result_previous = $this->get_rates_previous($i, $containers);

				if ($result_previous != null) {	
					$rates[] = $result_previous;
				} else {
					$cp_lookups++;
					if ($cp_lookups > $max_cp_lookups){
						// Error. Protect from too many lookups.
						$rates = null;
						break;
					}

					// Canada Post WebServices Lookup
					$result = $this->get_rates($country, $state, $postal, $shipping_weight, $length, $width, $height, $limit_services);
					$rates[] = $result;
				}
				// Save to container
				$containers[$i]['rate'] = $result;
				// Save distinct rate service_codes.
				$distinct_service_codes = $this->get_distinct_service_codes($distinct_service_codes, $result);
			}
			
			if (!empty($rates)){
				// Combine rates result sets into $rates_combined 
				
				if (count($rates) > 1 && count($rates[0]) > 0){
					// Loop through rates by type, combine to get total cost/type array().
					
					// Add first rate to $rates_combined.
					for($j=0;$j<count($rates[0]);$j++){
						$rates_combined[] = clone($rates[0][$j]);
					}
					
					// Now loop through the remainder of the rates array to create aggragate sum of prices.
					for($i=1;$i<count($rates);$i++){
						// Loop over objects.
						for($j=0;$j<count($rates[$i]);$j++){
							// Match using Service Code
							for($r=0; $r<count($rates_combined); $r++){
								if ($rates_combined[$r]->service_code == $rates[$i][$j]->service_code){
									// Sum Rate price
									$rates_combined[$r]->price = floatval($rates_combined[$r]->price) + floatval($rates[$i][$j]->price);
								}
							} // end foreach
						}// end for
					}// end for
					
					// Now loop through the $rates_combined array and only keep the rates that are in distinct_service_codes.
					if (!empty($distinct_service_codes)){
						for ($i=0;$i<count($rates_combined);$i++){
							// Check for valid service codes.
							if (!in_array($rates_combined[$i]->service_code, $distinct_service_codes)) {
								// oops, this service_code is not one of the distinct_service_codes. It has not been an aggragate sum so it needs to be removed.
								unset($rates_combined[$i]);
							}
						}
					}
					
				} else {
					// Only 1 rates result set
					$rates_combined = $rates[0];
				}
				
				// If services are the same cost, only keep the better service. (ie. Regular Parcel vs Epedited Parcel same cost)
				if ($this->options->prefer_service===true && count($rates_combined) > 1){
					$better_service = array_keys($this->available_services); // These are ordered by lowest to best already.
					$prev_cost = 0;
					$prev_service_code = '';
					// sort rates by lowest cost.
					usort($rates_combined, array(&$this, 'sort_rate_services'));
					// Loop from lowest to highest cost.
					for($i=0;$i<count($rates_combined);$i++){
						if ($rates_combined[$i]->price == $prev_cost){
							// Check their service_code position in the $better_service array.
							if (array_search($rates_combined[$i]->service_code, $better_service) > array_search($prev_service_code, $better_service)){
								// Remove the 'lower' service because it has the same cost.
								unset($rates_combined[$prev_index]);
							} else {
								unset($rates_combined[$i]);
								continue;
							}
						}
						$prev_cost = $rates_combined[$i]->price;
						$prev_service_code = $rates_combined[$i]->service_code;
						$prev_index = $i;
					}
					
				}
				
				// Do final foreach($rates_combined)
				
				foreach($rates_combined as $rate){
					if (!empty($this->options->margin) && $this->options->margin != 0) {
						 $rate->price = $rate->price * (1 + $this->options->margin/100); //Add margin
					}
					if (!empty($this->options->margin_value) && $this->options->margin_value != 0) {
						$rate->price = $rate->price + $this->options->margin_value; //Add margin_value
						if ($rate->price < 0){ $rate->price = 0; }
					}
					
					$box_margin = $this->get_box_margin_sum($containers);
					if ($box_margin > 0){
						$rate->price = $rate->price + $box_margin; //Add box margin if any value exists.
						if ($rate->price < 0){ $rate->price = 0; }
					}
					
					$delivery_label = '';
					if (!empty($this->options->delivery) && $rate->expected_delivery != '') {
						$delivery_label =  ' ('.__('Delivered by', 'woocommerce-canadapost-webservice') . ' ' . $rate->expected_delivery . ')';
						if (isset($rate->guaranteed) && ($rate->guaranteed == false) && isset($this->options->delivery_guarantee) && $this->options->delivery_guarantee) {
							$delivery_label = ''; // only display Delivery label on Guaranteed services (when $this->options->delivery_guarantee is enabled).
						}
					}
					
					$rateitem = array(
							'id' 		=> $this->id .':'. $rate->service_code,
							'label' 	=> $rate->service . $delivery_label,
							'cost' 		=> $rate->price
					);
					// Register the rate
					$this->add_rate( $rateitem );
					
				}
			} // endif 

			// Lettermail Limits.
			if ($this->options->lettermail_limits=='1' && !empty($this->options->lettermail_maxlength) && !empty($this->options->lettermail_maxwidth) && !empty($this->options->lettermail_maxheight)) {
				// Check to see if within lettermail limits.
				$lettermail_cubic =  $this->options->lettermail_maxlength * $this->options->lettermail_maxwidth * $this->options->lettermail_maxheight;
				if ($lettermail_cubic > 0) {
					if ($max['length'] <= $this->options->lettermail_maxlength && $max['width'] <= $this->options->lettermail_maxwidth && $max['height'] <= $this->options->lettermail_maxheight
					&& $cubic <= $lettermail_cubic) {
						// valid, within limit.
					} else {
						// over limit. Disable lettermail rates from being applied.
						$this->options->lettermail_enable = 0;
					}
				}
			}
				
			if ($this->options->lettermail_enable=='1'){
				/*
				 Canada Post Letter-post / Flat Rates
				*/
				// If override packing weight, remove package weight and add custom package weight.
				if (!empty($this->options->lettermail_override_weight) && $this->options->lettermail_override_weight){
					$shipping_weight -= (!empty($this->options->packageweight) ? floatval($this->options->packageweight) : 0);
					//$shipping_weight -= (!empty($box_fits['weight']) ? floatval($box_fits['weight']) : 0);
					$shipping_weight += (!empty($this->options->lettermail_packageweight) ? floatval($this->options->lettermail_packageweight) : 0);
					$shipping_weight = round($shipping_weight,2); // 2 decimal places.
				}
				// Subtotal for lettermail min/max subtotal calculation.
				$cart_subtotal = $woocommerce->cart->subtotal; 
				if ($this->options->lettermail_exclude_tax || true){  
					$tax = $woocommerce->cart->get_taxes_total(false, false); 
					// Remove tax from subtotal.
					if (is_numeric($tax)){ $cart_subtotal -= floatval($tax); }
				}
				
				foreach($this->lettermail as $lettermail) {
					if ($shipping_weight > $lettermail['weight_from'] && $shipping_weight < $lettermail['weight_to']
					&& ($country == $lettermail['country'] || ($lettermail['country']=='INT' && $country!='CA' && $country !='US'))
					&& (empty($lettermail['max_qty']) || $total_quantity <=  $lettermail['max_qty'])
					&& (empty($lettermail['min_total']) ||  $cart_subtotal >= $lettermail['min_total'])
					&& (empty($lettermail['max_total']) ||  $cart_subtotal <= $lettermail['max_total'])
					){
							
						$rateitem = array(
								'id' 		=> $this->id .':'.'Lettermail '.$lettermail['label'],
								'label' 	=> $lettermail['label'],
								'cost' 		=> $lettermail['cost']
						);
						$this->add_rate( $rateitem );
					}
				}

			}
			
			// Save shipping info to save with order to session.
			$shipping_info = array('rates'=>$rates_combined, 'packages'=>$containers);
			do_action('cpwebservice_order_shipping_info', $shipping_info);
				
			// Sort rates (by lowest cost)
			if(!empty($this->rates)){
				usort($this->rates, array(&$this, 'sort_rates'));
			}
		}
		// Logging
		if ( $this->options->log_enable ){
			$this->log->rates = $this->rates;
			$this->log->datestamp = current_time('timestamp');
			// Save to transient for 20 minutes.
			set_transient( 'cpwebservice_log', $this->log, 20 * MINUTE_IN_SECONDS );
		}

	}
		
	// Sort Rates function
	function sort_rates($a, $b){
		if ($a->cost == $b->cost) {
			return 0;
		}
		return ($a->cost < $b->cost) ? -1 : 1;
	}
	
	// Sort rate services function
	function sort_rate_services($a, $b){
		if ($a->price == $b->price) {
			return 0;
		}
		return ($a->price < $b->price) ? -1 : 1;
	}
	
	// Sort Boxes function. Ascending
	function sort_boxes($a, $b){
		$a_max = max($a['length'], $a['width'], $a['height']);
		$b_max = max($b['length'], $b['width'], $b['height']);
		if ($a['cubed'] == $b['cubed'] && $a_max == $b_max) {
			return 0;
		}
		if ($a['cubed'] == $b['cubed'] && $a_max < $b_max) {
			return -1;
		}
		return ($a['cubed'] < $b['cubed']) ? -1 : 1;
	}
	
	// Sort Products function.  Descending (biggest box first)
	function sort_products($a, $b){
		$a_max = max($a['length'], $a['width'], $a['height']);
		$b_max = max($b['length'], $b['width'], $b['height']);
		if ($a['cubic'] == $b['cubic'] && $a_max == $b_max) {
			return 0;
		}
		if ($a['cubic'] == $b['cubic'] && $a_max < $b_max) {
			return 1;
		}
		return ($a['cubic'] < $b['cubic']) ? 1 : -1;
	}
	
	// Rotates so that height is the smallest dimension. 
	// Also converts dimensions to float vals instead of strings.
	function rotate_package(&$length, &$width, &$height){
		$length = floatval($length);
		$width = floatval($width);
		$num = $height = floatval($height);
		if ($height > $length){
			$height = $length;
			$length = $num;
		}
		if ($height > $width){
			$height = $width;
			$width = $num;
		}
	}
	// Rotates so that for each box, height is the smallest dimension
	function rotate_boxes(&$boxes){
		for ($i=0;$i<count($boxes);$i++){
			$this->rotate_package($boxes[$i]['length'], $boxes[$i]['width'], $boxes[$i]['height']);
		}
	}
	
	// Gets distinct service codes. 
	// @unique array
	function get_distinct_service_codes($unique, $results){
		$service_codes = array();
		if (is_array($results)){
			for($i=0;$i<count($results);$i++){
				if (!in_array($results[$i]->service_code, $service_codes)){
					$service_codes[] = $results[$i]->service_code;
				}
			}
		}
		if (empty($unique)){ // if empty, this is the first $results set.
			return $service_codes;
		} else {
			// Go through $unique array and remove any service_code not found.
			for ($i=0;$i<count($unique);$i++){
				if (!in_array($unique[$i], $service_codes)){
					unset($unique[$i]);
				}
			}
			
			return $unique;
		}
	}
	
	function get_rates_previous($index, $containers){
		$rate = null;
		// after the 1st item.
		if ($index > 0){
			// Get Package/Container unique features.
			$weight_kg = $containers[$index]['weight'];
			$length = $containers[$index]['length'];
			$width = $containers[$index]['width'];
			$height = $containers[$index]['height'];
			$cubic = $containers[$index]['cubic'];

			// Look for previous container that is identical (and has rates -- after all, that's what we're looking for).
			for($j=0;$j<$index;$j++){
				if ($containers[$j]['weight'] == $weight_kg && $containers[$j]['length'] == $length && $containers[$j]['width'] == $width
		 		 &&  $containers[$j]['height'] == $height && $containers[$j]['cubic'] == $cubic && $containers[$j]['rate']!=null){
					// Huston, we have a match.
					return $containers[$j]['rate'];
				}
			}
			
		}
		return $rate;
	}
	
	function get_box_margin_sum($containers){
		$margin = 0;
		foreach($containers as $container){
			if (isset($container['box_margin']) && floatval($container['box_margin']) != 0){
				$margin += floatval($container['box_margin']);
			}
		}
		return $margin;
	}

	// Canada Post API rates lookup function
	function get_rates($dest_country, $dest_state, $dest_postal_code, $weight_kg, $length, $width, $height, $services = array()) {

		$rates = array();

		$username = $this->options->api_user;
		$password = $this->options->api_key;

		// REST URL
		$service_url = ($this->options->mode=='live') ? 'https://soa-gw.canadapost.ca/rs/ship/price' : 'https://ct.soa-gw.canadapost.ca/rs/ship/price'; // dev.  prod:

		// display errors
		$display_errors = ($this->options->display_errors ? true : false);

		// Has Services flag (Services are enabled for this country)
		$has_services = false;
		
		// if $services param is set, only request services within this array.
		$limit_services = (!empty($services) && is_array($services) && count($services) > 0);

		$xmlRequest =  new SimpleXMLElement(<<<XML
<?xml version="1.0" encoding="UTF-8"?>
<mailing-scenario xmlns="http://www.canadapost.ca/ws/ship/rate">
  <customer-number></customer-number>
  <parcel-characteristics>
    <weight></weight>
    <dimensions><length></length><width></width><height></height></dimensions>
  </parcel-characteristics>
  <services></services>
  <expected-mailing-date></expected-mailing-date>
  <origin-postal-code></origin-postal-code>
  <destination>
  </destination>
</mailing-scenario>
XML
		);

		// Create GetRates request xml
		$xmlRequest->{'origin-postal-code'} = $this->options->source_postalcode;
		$postalCode = str_replace(' ','',strtoupper($dest_postal_code)); //N0N0N0 (no spaces, uppercase)

		// Customer Number for venture one rates.
		if (!empty($this->options->account)){
			$xmlRequest->{'customer-number'} = $this->options->account;
			// Add Contract Id if entered.
			if (!empty($this->options->contractid)){
				$xmlRequest->addChild('contract-id', $this->options->contractid);
			}
		} else {
			// use public rates
			unset($xmlRequest->{'customer-number'});
			$xmlRequest->addChild('quote-type', 'counter');
		}


		// Add lead times.
		$lead_time = null;
		if (!empty($this->options->delivery)){
			$processTime = 86400 * $this->options->delivery; // in seconds
			$lead_time = date('Y-m-d', time() + $processTime);
			$xmlRequest->{'expected-mailing-date'} = $lead_time; 
		} else {
			unset($xmlRequest->{'expected-mailing-date'}); // remove element
		}

		// Parcel Dimensions
		if ($length > 0 || $width > 0 || $height > 0){

			// Round to 1 decimal place.
			$length = number_format(floatval($length),1,'.',''); $width = number_format(floatval($width),1,'.',''); $height = number_format(floatval($height),1,'.','');
				
			$xmlRequest->{'parcel-characteristics'}->dimensions->length = $length;
			$xmlRequest->{'parcel-characteristics'}->dimensions->width = $width;
			$xmlRequest->{'parcel-characteristics'}->dimensions->height = $height;
				
			//$dimensions = "<dimensions><length>{$length}</length><width>{$width}</width><height>{$height}</height></dimensions>";
		} else {
			unset($xmlRequest->{'parcel-characteristics'}->dimensions); // remove element
		}

		if ($dest_country == 'CA')
			// Canada
		{
			$xmlRequest->destination->addChild('domestic','');
			$xmlRequest->destination->domestic->addChild('postal-code', $postalCode);
			//$destination = "<domestic><postal-code>{$postalCode}</postal-code></domestic>";
				
			// Add Services
			foreach($this->services as $service_code){
				if (!empty($service_code) && strlen($service_code) >= 3 && substr($service_code,0,3) == 'DOM') {
					if (!$limit_services || in_array($service_code, $services)){
						$xmlRequest->services->addChild('service-code', $service_code);
						//$services .= "<service-code>$service_code</service-code>";
						$has_services = true;
					}
				}
			}
				
		} else if ($dest_country == 'US')
			// USA
		{
			$xmlRequest->destination->addChild('united-states','');
			$xmlRequest->destination->{'united-states'}->addChild('zip-code', $postalCode);
			//$destination = "<united-states><zip-code>{$postalCode}</zip-code></united-states>";
			// Add Services //$services = "<services><service-code>USA.EP</service-code><service-code>USA.SP.AIR</service-code><service-code>USA.XP</service-code></services>";
			foreach($this->services as $service_code){
				if (!empty($service_code) && strlen($service_code) >= 3 && substr($service_code,0,3) == 'USA') {
					if (!$limit_services || in_array($service_code, $services)){
						$xmlRequest->services->addChild('service-code', $service_code);
						$has_services = true;
					}
				}
			}
				
		} else
			// International
		{
			$xmlRequest->destination->addChild('international','');
			$xmlRequest->destination->international->addChild('country-code', $dest_country);
			//$destination = "<international><country-code>{$dest_country}</country-code></international>";
			// Add Services // $services = "<services><service-code>INT.XP</service-code><service-code>INT.IP.AIR</service-code><service-code>INT.SP.AIR</service-code></services>";
			foreach($this->services as $service_code){
				if (!empty($service_code) && strlen($service_code) >= 3 && substr($service_code,0,3) == 'INT') {
					if (!$limit_services || in_array($service_code, $services)){
						$xmlRequest->services->addChild('service-code', $service_code);
						$has_services = true;
					}
				}
			}
		}

		// Total Weight
		$xmlRequest->{'parcel-characteristics'}->weight = number_format( $weight_kg, 2, '.', '' ); // 2 decimal places : 99.99 format

		// Service Language: (English or French) sent as Accept-language header with a value of 'fr-CA' or 'en-CA'
		// If using WPML:
		if (defined('ICL_LANGUAGE_CODE')){
			$service_language = (ICL_LANGUAGE_CODE=='fr') ? 'fr-CA':'en-CA'; // 'en-CA' is default
		} else if (get_locale() == 'fr_FR'){
			$service_language = 'fr-CA';
		} else {
			$service_language = 'en-CA';
		}
		 
		$is_international = ($dest_country != 'US' && $dest_country != 'CA');
		
		// Zip/PostalCode Required.
		if ($this->options->display_required_notice && $has_services && empty($postalCode) && !$is_international && isset($_POST['calc_shipping_country'])) {
			if ( function_exists('wc_add_notice') ){  // 2.1 and higher
				wc_add_notice(__( 'Zip or Postal Code is required to calculate shipping.', 'woocommerce-canadapost-webservice' ), 'notice' );
			} else { // 2.0
				global $woocommerce;
				$woocommerce->add_message(__( 'Zip or Postal Code is required to calculate shipping.', 'woocommerce-canadapost-webservice' ) );
			}
		}
		 
		if ($has_services && ( !empty($username) && !empty($password) ) && (!empty($postalCode) || $is_international)){ // Postal code cannot be empty for CA or US.
			try {
				$curl = curl_init($service_url); // Create REST Request
				curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);
				curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, 2);
				curl_setopt($curl, CURLOPT_CAINFO, CPWEBSERVICE_PLUGIN_PATH . '/cert/cacert.pem'); // Signer Certificate in PEM format
				curl_setopt($curl, CURLOPT_POST, true);
				curl_setopt($curl, CURLOPT_POSTFIELDS, $xmlRequest->asXML());
				curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
				curl_setopt($curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
				curl_setopt($curl, CURLOPT_USERPWD, $username . ':' . $password);
				curl_setopt($curl, CURLOPT_HTTPHEADER, array('Content-Type: application/vnd.cpc.ship.rate+xml', 'Accept: application/vnd.cpc.ship.rate+xml','Accept-language: '.$service_language));
				$curl_response = curl_exec($curl); // Execute REST Request
				if(curl_errno($curl) && $display_errors){
					echo 'Curl error: ' . curl_error($curl) . "\n";
				}
				if ($display_errors || $this->options->log_enable) {
					$this->log->request['http'] = 'HTTP Response Status: ' . curl_getinfo($curl,CURLINFO_HTTP_CODE) . "\n";
					if ($display_errors){ echo $this->log->request['http']; }
				}
				curl_close($curl);

				// Using SimpleXML to parse xml response
				libxml_use_internal_errors(true);
				$xml = simplexml_load_string('<root>' . preg_replace('/<\?xml.*\?>/','',$curl_response) . '</root>');
				if (!$xml && ($display_errors || $this->options->log_enable)) {
					$errmsg = 'Failed loading XML' . "\n";
					$errmsg .= $curl_response . "\n";
					foreach(libxml_get_errors() as $error) {
						$errmsg.= "\t" . $error->message;
					}
					$this->log->request['errmsg'] = $errmsg;
					if ($display_errors) { echo $errmsg; }
				} else {
					if ($xml->{'price-quotes'} ) {
						$priceQuotes = $xml->{'price-quotes'}->children('http://www.canadapost.ca/ws/ship/rate');
						if ( $priceQuotes->{'price-quote'} ) {
							foreach ( $priceQuotes as $priceQuote ) {
								$rate = new stdClass();
								$rate->service =  (string)$priceQuote->{'service-name'};
								$rate->service_code =(string) $priceQuote->{'service-code'};
								$rate->price = number_format(floatval((string)$priceQuote->{'price-details'}->{'due'}),2,'.','');
								$rate->cubed = (string)$priceQuote->{'weight-details'}->{'cubed-weight'};
								$rate->guaranteed =(string)$priceQuote->{'service-standard'}->{'guaranteed-delivery'}; $rate->guaranteed = ($rate->guaranteed=='true');
								$rate->expected_delivery = (string) $priceQuote->{'service-standard'}->{'expected-delivery-date'};
								$rate->expected_mailing_date = $lead_time;
								// Add rate
								$rates[] = $rate;

								if ($display_errors || $this->options->log_enable) {
									$this->log->request['service'] .= "\nService: " . $priceQuote->{'service-name'} . "\n";
									$this->log->request['service'] .= 'Price: ' . $priceQuote->{'price-details'}->{'due'} . "\n";
									$this->log->request['service'] .= 'Expected Mailing Date: ' . $lead_time . "\n";
									$this->log->request['service'] .= 'Guaranteed Delivery: ' . $priceQuote->{'service-standard'}->{'guaranteed-delivery'} . "\n";
									$this->log->request['service'] .= 'Expected Delivery: ' . $priceQuote->{'service-standard'}->{'expected-delivery-date'} . "\n";
								}
							}
							if ($display_errors && isset($this->log->request['service'])){ echo $this->log->request['service']; }
						}
					}
					if ($xml->{'messages'} && ($display_errors || $this->options->log_enable)) {
						$apierror = '';
						$messages = $xml->{'messages'}->children('http://www.canadapost.ca/ws/messages');
						foreach ( $messages as $message ) {
							$apierror .= 'Error Code: ' . $message->code . "\n";
							$apierror .= 'Error Msg: ' . $message->description . "\n\n";
						}
						$this->log->request['apierror'] = $apierror;
						if ($display_errors){ echo $apierror; }
					}

				}
			} catch (Exception $ex) {
				// cURL request went wrong.
				if ($display_errors){
					echo 'Error: ' . $ex;
				}
				if ($this->options->log_enable){
					$this->log->request['error'] = 'Error: ' . $ex;
				}
			}
		} // endif $has_services

		return $rates;

	}
		

	function admin_options() {
		global $woocommerce;
		?>
				<?php // security nonce
					  wp_nonce_field(plugin_basename(__FILE__), 'cpwebservice_options_noncename'); 
				?>
				<h3><?php _e('Canada Post', 'woocommerce-canadapost-webservice'); ?></h3>
				<div><img src="<?php echo plugins_url( 'img/canada-post.jpg' , dirname(__FILE__) ); ?>" /></div>
				<table class="form-table">
					<tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Enable/Disable', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<fieldset><legend class="screen-reader-text"><span><?php _e('Enable/Disable', 'woocommerce-canadapost-webservice') ?></span></legend>
									<label for="woocommerce_cpwebservice_enabled">
									<input name="woocommerce_cpwebservice_enabled" id="woocommerce_cpwebservice_enabled" type="checkbox" value="1" <?php checked($this->options->enabled=='yes'); ?> /> <?php _e('Enable Canada Post Webservice', 'woocommerce-canadapost-webservice') ?></label><br>
								</fieldset>
						</td>
					    </tr>
					    <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Method Title', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<input type="text" name="woocommerce_cpwebservice_title" id="woocommerce_cpwebservice_title" style="min-width:50px;" value="<?php echo esc_attr($this->options->title); ?>" /> <span class="description"><?php _e('This controls the title which the user sees during checkout.', 'woocommerce-canadapost-webservice') ?></span>
						</td>
					    </tr>
					    <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Webservice Account Settings', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<input type="text" name="woocommerce_cpwebservice_account" id="woocommerce_cpwebservice_account" style="min-width:50px;" value="<?php echo esc_attr($this->options->account); ?>" /> <span class="description"><?php _e('Customer Number (Venture One)', 'woocommerce-canadapost-webservice') ?></span> <span id="woocommerce_cpwebservice_contractid_button" style="<?php echo (empty($this->options->contractid) ? "":"display:none"); ?>">
							<br />&nbsp; <a href="javscript:;" class="cpwebservice-showcontractid"><span class="description"><?php _e('Add Contract ID (Optional, Only if a Contract Customer)', 'woocommerce-canadapost-webservice')?></span></a>
							</span>
							<br />
							<div id="woocommerce_cpwebservice_contractid_display" style="<?php echo (!empty($this->options->contractid) ? "":"display:none"); ?>">
							<input type="text" name="woocommerce_cpwebservice_contractid" id="woocommerce_cpwebservice_contractid" style="min-width:50px;" value="<?php echo esc_attr($this->options->contractid); ?>" /> <span class="description"><?php _e('Contract ID (Optional, Only if a Contract Customer)', 'woocommerce-canadapost-webservice') ?></span>
							<br /></div>
							<input type="text" name="woocommerce_cpwebservice_api_user" id="woocommerce_cpwebservice_api_user" style="min-width:50px;" value="<?php echo esc_attr($this->options->api_user); ?>" /> <span class="description"><?php _e('API Username', 'woocommerce-canadapost-webservice') ?></span>
							<br />
							<input type="password" name="woocommerce_cpwebservice_api_key" id="woocommerce_cpwebservice_api_key" style="min-width:50px;" value="<?php echo esc_attr($this->options->api_key); ?>" /> <span class="description"><?php _e('API Password/Key', 'woocommerce-canadapost-webservice') ?></span>
							<br />
							<div><a href="<?php echo wp_nonce_url( admin_url( 'admin-ajax.php?action=cpwebservice_validate_api_credentials' ), 'cpwebservice_validate_api_credentials' ); ?>" id="woocommerce_cpwebservice_validate_btn" class="button cpwebservice-validate"><?php _e('Validate Credentials', 'woocommerce-canadapost-webservice') ?></a> <img src="<?php admin_url(); ?>/wp-admin/images/wpspin_light.gif" alt="" class="cpwebservice_ajaxupdate" style="display: none;" /><br /></div>							
							<div id="woocommerce_cpwebservice_validate" class="widefat" style="display:none;background-color: #fffbcc;padding:5px 0;border-color: #e6db55;"><p></p></div>
						</td>
					    </tr>
					    <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Origin Postal Code', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<input type="text" name="woocommerce_cpwebservice_source_postalcode" id="woocommerce_cpwebservice_source_postalcode" style="min-width:50px;" class="cpwebservice-postal" value="<?php echo esc_attr($this->options->source_postalcode); ?>" /> <span class="description"><?php _e('The Postal Code that items will be shipped from.', 'woocommerce-canadapost-webservice') ?></span>
							<div class="cpwebservice-postal-error" style="display:none;background-color: #fffbcc;padding:5px;border-color: #e6db55;"><p><?php _e('Warning: Postal Code is invalid. Required to be a valid Canadian postal code.', 'woocommerce-canadapost-webservice')?></p></div>
						</td>
					    </tr>
					    <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Order Shipping Tracking', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<label for="woocommerce_cpwebservice_shipping_tracking"><input name="woocommerce_cpwebservice_shipping_tracking" id="woocommerce_cpwebservice_shipping_tracking" type="checkbox" value="1" <?php checked($this->options->shipping_tracking==true); ?>  /> <?php _e('Enable Canada Post Tracking number feature on Orders', 'woocommerce-canadapost-webservice') ?></label>
							<p><label for="woocommerce_cpwebservice_email_tracking"><input name="woocommerce_cpwebservice_email_tracking" id="woocommerce_cpwebservice_email_tracking" type="checkbox" value="1" <?php checked($this->options->email_tracking==true); ?>  /> <?php _e('Enable Email notification when Parcel Tracking updates', 'woocommerce-canadapost-webservice') ?></label> <span class="description"><?php _e('Automatic email notifications to customers when "Mailed on" or "Delivered" date is updated', 'woocommerce-canadapost-webservice')?></span></p>
							<p><label for="woocommerce_cpwebservice_tracking_icons"><input name="woocommerce_cpwebservice_tracking_icons" id="woocommerce_cpwebservice_tracking_icons" type="checkbox" value="1" <?php checked($this->options->tracking_icons==true); ?>  /> <?php _e('Display icons with Tracking information', 'woocommerce-canadapost-webservice') ?></label>
						</td>
					    </tr>
					    <tr valign="top">
					    <th scope="row" class="titledesc"><?php _e('Development', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
							<fieldset><legend class="screen-reader-text"><span><?php _e('Development Mode', 'woocommerce-canadapost-webservice') ?></span></legend>
									<select name="woocommerce_cpwebservice_mode">
										<option value="dev"<?php if ($this->options->mode=='dev') echo 'selected="selected"'; ?>>Development</option>
										<option value="live" <?php if ($this->options->mode=='live') echo 'selected="selected"'; ?>>Production/Live</option>
									</select>
									<br />
									<label for="woocommerce_cpwebservice_display_errors">
									<input name="woocommerce_cpwebservice_display_errors" id="woocommerce_cpwebservice_display_errors" type="checkbox" value="1" <?php checked($this->options->display_errors=='1'); ?> /> <?php _e('Display Errors', 'woocommerce-canadapost-webservice') ?> <small><?php _e('Warning: Do not enable in production since debug/errors would be shown to customers', 'woocommerce-canadapost-webservice') ?></small></label><br>
									</fieldset>
						</td>
					    </tr>
					    <tr>
						<th scope="row" class="titledesc"><?php _e('Add Margin', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
								&nbsp; <input type="text" name="woocommerce_cpwebservice_margin" id="woocommerce_cpwebservice_margin" style="max-width:50px;" value="<?php echo esc_attr($this->options->margin); ?>" />% <span class="description"><?php _e('Add Margin Percentage (ex. 5% or -2%) to Shipping Cost', 'woocommerce-canadapost-webservice') ?></span><br />
								$<input type="text" name="woocommerce_cpwebservice_margin_value" id="woocommerce_cpwebservice_margin_value" style="max-width:50px;" value="<?php echo esc_attr($this->options->margin_value); ?>" /> &nbsp; <span class="description"><?php _e('Add Margin Amount (ex. $4 or -$1) to Shipping Cost', 'woocommerce-canadapost-webservice') ?></span>
						</td>
					    </tr>
					    <tr>
						<th scope="row" class="titledesc"><?php _e('Box/Envelope Weight', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
								<input type="text" name="woocommerce_cpwebservice_packageweight" id="woocommerce_cpwebservice_packageweight" style="max-width:50px;" value="<?php echo esc_attr($this->options->packageweight); ?>" />kg <span class="description"><?php _e('Envelope/Box weight with bill/notes/advertising inserts (ex. 0.02kg)', 'woocommerce-canadapost-webservice') ?></span>
						</td>
					    </tr>
					    <tr>
					    <th scope="row" class="titledesc"><?php _e('Box Packing', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
								
								<p><label for="woocommerce_cpwebservice_box_packing">
								<input name="woocommerce_cpwebservice_box_packing" id="woocommerce_cpwebservice_box_packing" type="checkbox" value="1" checked="checked" disabled="disabled" /> <?php _e('Advanced Box-Packing Algorithm used to pack products.', 'woocommerce-canadapost-webservice') ?></label>
								</p>
								<p>								
								<label for="woocommerce_cpwebservice_volumetric_weight">
								<input name="woocommerce_cpwebservice_volumetric_weight" id="woocommerce_cpwebservice_volumetric_weight" type="checkbox" value="1" <?php checked(!empty($this->options->volumetric_weight)); ?> /> <?php _e('After packing, use Volumetric weight (if it is more than the actual weight) when requesting package rates (Recommended by Canada Post)', 'woocommerce-canadapost-webservice') ?></label>
								<img class="help_tip" data-tip="<?php esc_attr_e( 'Most shipping couriers, including Canada Post charge by volumetric weight (the size of the package) when it is greater than the actual weight.', 'woocommerce' ); ?>" src="<?php echo esc_url( WC()->plugin_url() ); ?>/assets/images/help.png" height="16" width="16" /></p>
								
								<p><label for="woocommerce_cpwebservice_weight_only_enabled">
								<input name="woocommerce_cpwebservice_weight_only_enabled" id="woocommerce_cpwebservice_weight_only_enabled" type="checkbox" value="1" <?php checked(!empty($this->options->weight_only_enabled)); ?> /> <?php _e('Allow products with weight-only (no dimensions) to still be calculated.', 'woocommerce-canadapost-webservice') ?></label>
								</p>
								<p>
								<label for="woocommerce_cpwebservice_product_shipping_options">
								<input name="woocommerce_cpwebservice_product_shipping_options" id="woocommerce_cpwebservice_product_shipping_options" type="checkbox" value="1" <?php checked(!empty($this->options->product_shipping_options)); ?> /> <?php _e('Enable the option to mark products as Ready-to-Ship (Pre-packaged)', 'woocommerce-canadapost-webservice') ?></label>
								<br />
								<span class="description"><?php _e('This option will be displayed on the Product Edit page in Woocommerce.', 'woocommerce-canadapost-webservice') ?></span>
								</p>
								
						</td>
					    </tr>
					    <tr>
						<th scope="row" class="titledesc"><?php _e('Delivery Dates', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
									<p><label for="woocommerce_cpwebservice_delivery_hide">
									<input name="woocommerce_cpwebservice_delivery_hide" id="woocommerce_cpwebservice_delivery_hide" onclick="jQuery('#woocommerce_cpwebservice_delivery').val(this.checked?'1':'');" type="checkbox" value="1" <?php checked(!empty($this->options->delivery)); ?> /> <?php _e('Enable Estimated Delivery Dates', 'woocommerce-canadapost-webservice') ?></label>
									</p>
									<p><label for="woocommerce_cpwebservice_delivery_guarantee"><input name="woocommerce_cpwebservice_delivery_guarantee" id="woocommerce_cpwebservice_delivery_guarantee" type="checkbox" value="1" <?php checked(!empty($this->options->delivery_guarantee)); ?> /> <?php _e('Show Estimated Delivery only on Guaranteed Services', 'woocommerce-canadapost-webservice') ?></label> <span class="description"><?php _e('(ie. Regular Parcel will not display, only Xpresspost, Priority, etc)', 'woocommerce-canadapost-webservice') ?></span></p>
									<p><input type="text" name="woocommerce_cpwebservice_delivery" id="woocommerce_cpwebservice_delivery" style="max-width:50px;" value="<?php echo esc_attr($this->options->delivery); ?>" /> <span class="description"><?php _e('Days to Ship after order placed.  Used to calculate date mailed.', 'woocommerce-canadapost-webservice') ?></span></p>
						</td>
					    </tr>
					    <tr>
					    <th scope="row" class="titledesc"><?php _e('Validation', 'woocommerce-canadapost-webservice')?> </th>
						<td class="forminp">
								<label for="woocommerce_cpwebservice_display_required_notice">
									<input name="woocommerce_cpwebservice_display_required_notice" id="woocommerce_cpwebservice_display_required_notice" type="checkbox" value="1" <?php checked(!empty($this->options->display_required_notice)); ?> /> <?php _e('Validate Zip / Postal code as required in Calculate Shipping form', 'woocommerce-canadapost-webservice') ?></label>
						</td>
					    </tr>
					    
					    <tr>
						<th scope="row" class="titledesc"><?php _e('Enable Parcel Services', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
								<fieldset><legend class="screen-reader-text"><span><?php _e('Canada Post Shipping Services', 'woocommerce-canadapost-webservice') ?></span></legend>
								<?php if (empty($this->services)){ $this->services = array_keys($this->available_services);  } // set all checked as default.
									  $s=0; // service count
									  $cur_code = ''; // heading label ?>
								<?php foreach($this->available_services as $service_code=>$service_label): ?>
								<?php $s++;
									  if (!empty($service_code) && strlen($service_code) >= 3){ $prefix = substr($service_code,0,3); } else { $prefix = ''; }
								      if ($cur_code!=$prefix){ $cur_code=$prefix; echo '<h4>'.esc_html($this->get_destination_from_service($service_code)).'</h4>'; } ?>
									<label for="woocommerce_cpwebservice_service-<?php echo $s ?>">
									<input name="woocommerce_cpwebservice_services[]" id="woocommerce_cpwebservice_service-<?php echo $s ?>" type="checkbox" value="<?php echo esc_attr($service_code) ?>" <?php checked(in_array($service_code,$this->services)); ?> /> <?php echo esc_html($service_label); ?></label><br>
								<?php endforeach; ?>
								</fieldset>
						</td>
					    </tr>
					    <tr>
					    <th scope="row" class="titledesc"><?php _e('Parcel Services', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
								<p><label for="woocommerce_cpwebservice_prefer_service">
								<input name="woocommerce_cpwebservice_prefer_service" id="woocommerce_cpwebservice_prefer_service" type="checkbox" value="1" <?php checked(!empty($this->options->prefer_service)); ?> /> <?php _e('If services are the same cost, only display the better service.', 'woocommerce-canadapost-webservice') ?></label>
								<br />
								<span class="description"><?php _e('(ie. if Regular Parcel and Expedited Parcel are same cost, keep Expedited Parcel Service.)', 'woocommerce-canadapost-webservice')?></span>
								</p>
						</td>
					    </tr>
					    <tr>
					    <th scope="row" class="titledesc"><?php _e('Shipping Class Rules', 'woocommerce-canadapost-webservice')?> </th>
						<td class="forminp">
						
								<p><label for="woocommerce_cpwebservice_enable_rules">
									<input name="woocommerce_cpwebservice_enable_rules" id="woocommerce_cpwebservice_enable_rules" type="checkbox" value="1" <?php checked(!empty($this->options->rules_enable)); ?> /> <?php _e('Enable Shipping Class rules', 'woocommerce-canadapost-webservice') ?></label> &nbsp; <span class="description"><?php _e('Assign products to these shipping classes to apply these rules.', 'woocommerce-canadapost-webservice')?></span></p>
						<div id="cpwebservice_rules">									
								<?php $this->shipping_class_rule(); ?>
						</div>
						</td>
					    </tr>
					    
					    <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Shipping Package/Box sizes', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
						<?php if (!$this->boxes || !is_array($this->boxes) || count($this->boxes) == 0){
								$this->boxes = array(array('length'=>'9','width'=>'6', 'height'=>'15','name'=>'Standard Box'));
								$this->options->boxes_enable='1';
							}
							$box_defaults = array('length'=>0,'width'=>0, 'height'=>0,'name'=>''); ?>
						<label for="woocommerce_cpwebservice_boxes_enable">
									<input name="woocommerce_cpwebservice_boxes_enable" id="woocommerce_cpwebservice_boxes_enable" type="checkbox" value="1" <?php checked($this->options->boxes_enable=='1'); ?> /> <?php _e('Enable Shipping Box Sizes', 'woocommerce-canadapost-webservice') ?></label><br />
							<span class="description"><?php _e('Please define a number of envelope/package/box sizes that you use to ship. Dimensions are in cm.  These will be used to ship but if a large enough box is not found for a product, the system will use a calculate one.', 'woocommerce-canadapost-webservice') ?></span>
							<div id="cpwebservice_boxes">							
								<?php for($i=0;$i<count($this->boxes); $i++): ?>
								<?php $box = (is_array($this->boxes[$i]) ? array_merge($box_defaults, $this->boxes[$i]) : array()); ?>
								<p class="form-field">
								<label for="woocommerce_cpwebservice_box_length[]"><?php _e('Box Dimensions (cm)', 'woocommerce-canadapost-webservice'); ?></label><span class="wrap">
										<input name="woocommerce_cpwebservice_box_length[]" id="woocommerce_cpwebservice_box_length<?php echo $i;?>" style="max-width:50px" placeholder="Length" class="input-text" size="6" type="text" value="<?php echo esc_attr($box['length']); ?>" />
										<input name="woocommerce_cpwebservice_box_width[]" id="woocommerce_cpwebservice_box_width<?php echo $i;?>" style="max-width:50px" placeholder="Width" class="input-text" size="6" type="text" value="<?php echo esc_attr($box['width']); ?>">
										<input name="woocommerce_cpwebservice_box_height[]" id="woocommerce_cpwebservice_box_width<?php echo $i;?>" style="max-width:50px" placeholder="Height" class="input-text last" size="6" type="text" value="<?php echo esc_attr($box['height']); ?>" />
										<span class="description"><?php _e('LxWxH cm decimal form','woocommerce-canadapost-webservice'); ?></span>
										<span class="description" style="margin-left:10px;"><?php _e('Box Name (internal)', 'woocommerce-canadapost-webservice'); ?>:</span>
										<input name="woocommerce_cpwebservice_box_name[]" id="woocommerce_cpwebservice_box_name<?php echo $i;?>" style="max-width:120px;" placeholder="Box Name" class="input-text last" size="6" type="text" value="<?php echo esc_attr($box['name']); ?>"></span>
										<span class="description"> <?php _e('Box Weight','woocommerce-canadapost-webservice')?>:</span><input name="woocommerce_cpwebservice_box_weight[]" id="woocommerce_cpwebservice_box_weight<?php echo $i;?>" style="max-width:50px" class="input-text last" size="6" type="text" value="<?php echo esc_attr(isset($box['weight'])? $box['weight'] : ''); ?>" /><span class="description">kg</span>
										<span class="description">&nbsp; <?php _e('Add Cost/Margin $','woocommerce-canadapost-webservice')?></span><input name="woocommerce_cpwebservice_box_margin[]" id="woocommerce_cpwebservice_box_margin<?php echo $i;?>" style="max-width:50px" class="input-text last" size="6" type="text" value="<?php echo esc_attr(isset($box['margin'])? $box['margin'] : ''); ?>" />
										<span style="margin-left:5px;"><a href="javascript:;" title="Remove" onclick="jQuery(this).parent().parent('p').remove();" class="button"><?php _e('Remove','woocommerce-canadapost-webservice'); ?></a></span>
								</p>
								<?php endfor; ?>
							</div>
							<a href="javascript:;" id="btn_cpwebservice_boxes" class="button-secondary"><?php _e('Add More +','woocommerce-canadapost-webservice'); ?></a>
						</td>
					    </tr>
					    
					     <tr valign="top">
						<th scope="row" class="titledesc"><?php _e('Lettermail / Flat Rates', 'woocommerce-canadapost-webservice') ?></th>
						<td class="forminp">
						<?php if (empty($this->lettermail) || !is_array($this->lettermail) || count($this->lettermail) == 0){
								// Set default CP Lettermail Rates.
								/*Canada Post Letter-post USA rates:(for now)
								0-100g = $2.20
								100g-200g = $3.80
								200g-500g = $7.60*/
								$this->lettermail = array(array('country'=>'CA','label'=>'Canada Post Lettermail', 'cost'=>'2.20','weight_from'=>'0','weight_to'=>'0.1', 'max_qty'=>0, 'min_total'=>'', 'max_total'=>''),
													    array('country'=>'CA','label'=>'Canada Post Lettermail', 'cost'=>'3.80','weight_from'=>'0.1','weight_to'=>'0.2', 'max_qty'=>0, 'min_total'=>'', 'max_total'=>''),
														array('country'=>'US','label'=>'Canada Post Lettermail', 'cost'=>'2.20','weight_from'=>'0','weight_to'=>'0.1', 'max_qty'=>0, 'min_total'=>'', 'max_total'=>''),
														array('country'=>'US','label'=>'Canada Post Lettermail', 'cost'=>'3.80','weight_from'=>'0.1','weight_to'=>'0.2', 'max_qty'=>0, 'min_total'=>'', 'max_total'=>''),
														array('country'=>'US','label'=>'Canada Post Lettermail', 'cost'=>'7.60','weight_from'=>'0.2','weight_to'=>'0.5', 'max_qty'=>0, 'min_total'=>'', 'max_total'=>''));
								$this->options->lettermail_enable='';
							} 
							$lettermail_defaults = array('country'=>'','label'=>'', 'cost'=>0,'weight_from'=>'','weight_to'=>'','max_qty'=>0, 'min_total'=>'', 'max_total'=>''); ?>
						<label for="woocommerce_cpwebservice_lettermail_enable">
									<input name="woocommerce_cpwebservice_lettermail_enable" id="woocommerce_cpwebservice_lettermail_enable" type="checkbox" value="1" <?php checked($this->options->lettermail_enable=='1'); ?> /> <?php _e('Enable Lettermail / Flat Rates', 'woocommerce-canadapost-webservice') ?></label>
							<p class="description"><?php _e('Define Lettermail rates based on Weight Range (kg)', 'woocommerce-canadapost-webservice') ?>.</p>
							<p class="description"> <?php _e('Example: 0.1kg to 0.2kg: $3.80 Lettermail', 'woocommerce-canadapost-webservice') ?></p>
							<div id="cpwebservice_lettermail">							
								<?php for($i=0;$i<count($this->lettermail); $i++): ?>
								<?php $lettermail = (is_array($this->lettermail[$i]) ? array_merge($lettermail_defaults, $this->lettermail[$i]) : array()); ?>
								<p class="form-field">
								
								<span class="wrap">
								    <select name="woocommerce_cpwebservice_lettermail_country[]">
										<option value="CA"<?php if ($lettermail['country']=='CA') echo 'selected="selected"'; ?>>Canada</option>
										<option value="US" <?php if ($lettermail['country']=='US') echo 'selected="selected"'; ?>>USA</option>
										<option value="INT" <?php if ($lettermail['country']=='INT') echo 'selected="selected"'; ?>><?php _e('International', 'woocommerce-canadapost-webservice') ?></option>
									</select>
									<label for="woocommerce_cpwebservice_lettermail_label<?php echo $i;?>"> <?php _e('Label', 'woocommerce-canadapost-webservice'); ?>:</label>
										<input name="woocommerce_cpwebservice_lettermail_label[]" id="woocommerce_cpwebservice_lettermail_label<?php echo $i;?>" style="max-width:150px" placeholder="<?php _e('Lettermail','woocommerce-canadapost-webservice'); ?>" class="input-text" size="16" type="text" value="<?php echo esc_attr($lettermail['label']); ?>" />
										<span class="description"> <?php _e('Cost','woocommerce-canadapost-webservice'); ?>: $</span><input name="woocommerce_cpwebservice_lettermail_cost[]" id="woocommerce_cpwebservice_lettermail_cost<?php echo $i;?>" style="max-width:50px" placeholder="<?php _e('Cost','woocommerce-canadapost-webservice'); ?>" class="input-text" size="16" type="text" value="<?php echo esc_attr($lettermail['cost']); ?>">
										<span class="description"> <?php _e('Weight Range(kg)','woocommerce-canadapost-webservice'); ?>: </span><input name="woocommerce_cpwebservice_lettermail_weight_from[]" id="woocommerce_cpwebservice_lettermail_weight_from<?php echo $i;?>" style="max-width:40px" placeholder="" class="input-text" size="6" type="text" value="<?php echo esc_attr($lettermail['weight_from']); ?>" />kg
										 <?php _e('to','woocommerce-canadapost-webservice'); ?> &lt;
										<input name="woocommerce_cpwebservice_lettermail_weight_to[]" id="woocommerce_cpwebservice_lettermail_weight_to<?php echo $i;?>" style="max-width:40px" placeholder="" class="input-text last" size="6" type="text" value="<?php echo esc_attr($lettermail['weight_to']); ?>" />kg</span>
										<span class="description"> <?php _e('Max items (0 for no limit)','woocommerce-canadapost-webservice'); ?>: </span> <input name="woocommerce_cpwebservice_lettermail_max_qty[]" id="woocommerce_cpwebservice_lettermail_max_qty<?php echo $i;?>" style="max-width:40px" placeholder="" class="input-text" size="6" type="text" value="<?php echo esc_attr($lettermail['max_qty']); ?>" />
										<span class="description"> <?php _e('Cart subtotal','woocommerce-canadapost-webservice'); ?>: $</span><input name="woocommerce_cpwebservice_lettermail_min_total[]" id="woocommerce_cpwebservice_lettermail_min_total<?php echo $i;?>" style="max-width:50px" placeholder="" class="input-text" size="6" type="text" value="<?php echo esc_attr($lettermail['min_total']); ?>" />
										 <?php _e('to','woocommerce-canadapost-webservice'); ?> &lt;=
										$<input name="woocommerce_cpwebservice_lettermail_max_total[]" id="woocommerce_cpwebservice_lettermail_max_total<?php echo $i;?>" style="max-width:50px" placeholder="" class="input-text last" size="6" type="text" value="<?php echo esc_attr($lettermail['max_total']); ?>" />
										<span style="margin-left:5px;"><a href="javascript:;" title="Remove" onclick="jQuery(this).parent().parent('p').remove();" class="button"><?php _e('Remove','woocommerce-canadapost-webservice'); ?></a></span>
								</p>
								<?php endfor; ?>
							</div>
							<a href="javascript:;" id="btn_cpwebservice_lettermail" class="button-secondary"><?php _e('Add More +','woocommerce-canadapost-webservice'); ?></a>
							<br />
							<br />
							<label for="woocommerce_cpwebservice_lettermail_limits">
									<input name="woocommerce_cpwebservice_lettermail_limits" id="woocommerce_cpwebservice_lettermail_limits" type="checkbox" value="1" <?php checked($this->options->lettermail_limits=='1'); ?> /> <?php _e('Maximum dimensions for Lettermail/Flat Rates (Also maximum volumetric weight)', 'woocommerce-canadapost-webservice') ?></label>
							<p class="form-field">
								<input name="woocommerce_cpwebservice_lettermail_maxlength" id="woocommerce_cpwebservice_lettermail_maxlength" style="max-width:50px" placeholder="Length" class="input-text" size="6" type="text" value="<?php echo esc_attr($this->options->lettermail_maxlength); ?>" />
										<input name="woocommerce_cpwebservice_lettermail_maxwidth" id="woocommerce_cpwebservice_lettermail_maxwidth" style="max-width:50px" placeholder="Width" class="input-text" size="6" type="text" value="<?php echo esc_attr($this->options->lettermail_maxwidth); ?>">
										<input name="woocommerce_cpwebservice_lettermail_maxheight" id="woocommerce_cpwebservice_lettermail_maxheight" style="max-width:50px" placeholder="Height" class="input-text last" size="6" type="text" value="<?php echo esc_attr($this->options->lettermail_maxheight); ?>" />
										<span class="description"><?php _e('LxWxH cm decimal form','woocommerce-canadapost-webservice'); ?></span>
							</p>
							<br />
							<label for="woocommerce_cpwebservice_lettermail_override_weight">
									<input name="woocommerce_cpwebservice_lettermail_override_weight" id="woocommerce_cpwebservice_lettermail_override_weight" type="checkbox" value="1" <?php checked($this->options->lettermail_override_weight=='1'); ?> /> <?php _e('Override Box/Envelope Weights for Lettermail/Flat Rates', 'woocommerce-canadapost-webservice') ?></label>
							<p class="form-field">
								<input name="woocommerce_cpwebservice_lettermail_packageweight" id="woocommerce_cpwebservice_lettermail_packageweight" style="max-width:50px" class="input-text" size="6" type="text" value="<?php echo esc_attr($this->options->lettermail_packageweight); ?>" />kg <span class="description"><?php _e('Envelope/Box weight. This is used instead of above Box/Envelope Weight, but only for calculating Lettermail/Flat Rates.', 'woocommerce-canadapost-webservice') ?></span></p>
						</td>
					    </tr>
					    
				</table>
				<script type="text/javascript">
					jQuery(document).ready(function($) {

						$('.cpwebservice-showcontractid').click(function(){
							jQuery('#woocommerce_cpwebservice_contractid_button').hide();
							jQuery('#woocommerce_cpwebservice_contractid_display').show();
						});

						$('.cpwebservice-validate').click(function(){
							var url = $(this).attr('href');
							var postvalues= { api_user:$('input#woocommerce_cpwebservice_api_user').val(),
									api_key:$('input#woocommerce_cpwebservice_api_key').val(),
									customerid:$('input#woocommerce_cpwebservice_account').val(),
									contractid:$('input#woocommerce_cpwebservice_contractid').val() };
							// ajax request.
							$('img.cpwebservice_ajaxupdate').show();
							$.post(url,postvalues,function(data){
								console.log('Data:'+data);
								$('#woocommerce_cpwebservice_validate p').html(data);
								$('#woocommerce_cpwebservice_validate').show();
								$('img.cpwebservice_ajaxupdate').hide();
							});
							return false;
						});

						$('.cpwebservice-log-display').click(function(){
							var url = $(this).attr('href');
							$('img.cpwebservice-log-display-loading').show();
							$('#cpwebservice_log_display').hide();
							$.get(url,function(data){
								//console.log('Data:'+data);
								$('#cpwebservice_log_display').html(data);
								$('#cpwebservice_log_display').slideDown();
								$('img.cpwebservice-log-display-loading').hide();
							});
							return false;
						});

						// Validate Postal Code
						var validPostalCode = function() {
							var regex = /^[A-Za-z]{1}\d{1}[A-Za-z]{1} *\d{1}[A-Za-z]{1}\d{1}$/;
							$('.cpwebservice-postal-error').toggle(regex.test($('.cpwebservice-postal').val()) == false);
						}
						if ($('.cpwebservice-postal').val() != '') { validPostalCode(); }
						$('.cpwebservice-postal').on('blur', validPostalCode);
						
						
						jQuery('#btn_cpwebservice_boxes').click(function() {
						
							var i = jQuery('#cpwebservice_boxes p').size(); // one p tag.
							
							var fields = '';
							fields += '<p class="form-field">';
							fields += '<label for="woocommerce_cpwebservice_box_length[]"><?php _e('Box Dimensions (cm)', 'woocommerce-canadapost-webservice'); ?></label><span class="wrap"> ';
							fields += '<input name="woocommerce_cpwebservice_box_length[]" id="woocommerce_cpwebservice_box_length'+i+'" style="max-width:50px" placeholder="Length" class="input-text" size="6" type="text" value="" />';
							fields += '<input name="woocommerce_cpwebservice_box_width[]" id="woocommerce_cpwebservice_box_width'+i+'" style="max-width:50px" placeholder="Width" class="input-text" size="6" type="text" value="">';
							fields += '<input name="woocommerce_cpwebservice_box_height[]" id="woocommerce_cpwebservice_box_width'+i+'" style="max-width:50px" placeholder="Height" class="input-text" size="6" type="text" value="" />';
							fields += '<span class="description"><?php _e('LxWxH cm decimal form','woocommerce-canadapost-webservice'); ?></span> ';
							fields += '<span class="description" style="margin-left:10px;"><?php _e('Box Name (internal)', 'woocommerce-canadapost-webservice'); ?>:</span>';
							fields += '<input name="woocommerce_cpwebservice_box_name[]" id="woocommerce_cpwebservice_box_name'+i+'" style="max-width:120px;" placeholder="Box Name" class="input-text last" size="6" type="text" value=""></span>';
							fields += '<span class="description"> <?php _e('Box Weight','woocommerce-canadapost-webservice')?>:</span><input name="woocommerce_cpwebservice_box_weight[]" id="woocommerce_cpwebservice_box_weight'+i+'" style="max-width:50px" class="input-text last" size="6" type="text" value="" /><span class="description">kg</span>';
							fields += '<span class="description">&nbsp; <?php _e('Add Cost/Margin $','woocommerce-canadapost-webservice')?></span><input name="woocommerce_cpwebservice_box_margin[]" id="woocommerce_cpwebservice_box_margin'+i+'" style="max-width:50px" class="input-text last" size="6" type="text" value="" />';
							fields += '<span style="margin-left:5px;"><a href="javascript:;" title="Remove" onclick="'+"jQuery(this).parent().parent('p').remove();"+'" class="button"><?php _e('Remove','woocommerce-canadapost-webservice'); ?></a></span>';
							fields += '</p>';
							
							jQuery(fields).appendTo('#cpwebservice_boxes');
						});

						jQuery('#btn_cpwebservice_lettermail').click(function() {
							
							var i = jQuery('#cpwebservice_lettermail p').size(); // one p tag.
							
							var fields = '';
							fields += '<p class="form-field">';
							fields += '<select name="woocommerce_cpwebservice_lettermail_country[]"><option value="CA" selected="selected">Canada</option><option value="US">USA</option><option value="INT"><?php _e('International', 'woocommerce-canadapost-webservice') ?></option></select>';
							fields += '<label> <?php _e('Label', 'woocommerce-canadapost-webservice'); ?>:</label>';
							fields += '<input name="woocommerce_cpwebservice_lettermail_label[]" id="woocommerce_cpwebservice_lettermail_label'+i+'" style="max-width:150px" placeholder="<?php _e('Lettermail','woocommerce-canadapost-webservice'); ?>" class="input-text" size="16" type="text" value="">';
							fields += '<span class="description"> <?php _e('Cost','woocommerce-canadapost-webservice'); ?>: $</span><input name="woocommerce_cpwebservice_lettermail_cost[]" id="woocommerce_cpwebservice_lettermail_cost'+i+'" style="max-width:50px" placeholder="<?php _e('Cost','woocommerce-canadapost-webservice'); ?>" class="input-text" size="16" type="text" value="">';
							fields += '<span class="description"> <?php _e('Weight Range(kg)','woocommerce-canadapost-webservice'); ?>: </span><input name="woocommerce_cpwebservice_lettermail_weight_from[]" id="woocommerce_cpwebservice_lettermail_weight_from'+i+'" style="max-width:40px" placeholder="" class="input-text" size="6" type="text" value="0">kg';
							fields += ' <?php _e('to','woocommerce-canadapost-webservice'); ?> &lt;=';
							fields += '<input name="woocommerce_cpwebservice_lettermail_weight_to[]" id="woocommerce_cpwebservice_lettermail_weight_to'+i+'" style="max-width:40px" placeholder="" class="input-text last" size="6" type="text" value="0">kg</span>';
							fields += '<span class="description"> <?php _e('Max items (0 for no limit)','woocommerce-canadapost-webservice'); ?>: </span> <input name="woocommerce_cpwebservice_lettermail_max_qty[]" id="woocommerce_cpwebservice_lettermail_max_qty'+i+'" style="max-width:40px" placeholder="" class="input-text" size="6" type="text" value="" />';
							fields += '<span class="description"> <?php _e('Cart subtotal','woocommerce-canadapost-webservice'); ?>: $</span><input name="woocommerce_cpwebservice_lettermail_min_total[]" id="woocommerce_cpwebservice_lettermail_min_total'+i+'" style="max-width:50px" placeholder="" class="input-text" size="6" type="text" value="" />';
							fields += '<?php _e('to','woocommerce-canadapost-webservice'); ?> &lt; $<input name="woocommerce_cpwebservice_lettermail_max_total[]" id="woocommerce_cpwebservice_lettermail_max_total'+i+'" style="max-width:50px" placeholder="" class="input-text last" size="6" type="text" value="" />';
							fields += '<span style="margin-left:5px;"><a href="javascript:;" title="Remove" onclick="'+"jQuery(this).parent().parent('p').remove();"+'" class="button"><?php _e('Remove','woocommerce-canadapost-webservice'); ?></a></span>';
							fields += '</p>';
							
							jQuery(fields).appendTo('#cpwebservice_lettermail');
						});

						jQuery('.btn_cpwebservice_rules').on('click',function() {
							var fields = jQuery(this).parent().parent().find('select').each(function() { jQuery(this).find('option:selected').removeAttr('selected'); });
							<?php if ( defined( 'WOOCOMMERCE_VERSION' ) && version_compare( WOOCOMMERCE_VERSION, '2.3', '>=' ) ) { ?>
							fields.trigger('change');
							<?php } else { // chosen ?>
							fields.trigger('chosen:updated');
							<?php } // endif;?>
						});
					});

					</script>
					
					<table class="form-table">
					<tr valign="top">
					    <th scope="row" class="titledesc">
					    	<?php _e('Logging', 'woocommerce-canadapost-webservice')?>
					    </th>
						<td class="forminp">
						<label for="woocommerce_cpwebservice_log_enable">
									<input name="woocommerce_cpwebservice_log_enable" id="woocommerce_cpwebservice_log_enable" type="checkbox" value="1" <?php checked($this->options->log_enable=='1'); ?> /> <?php _e('Enable Rates Lookup Logging', 'woocommerce-canadapost-webservice') ?>
									<br /><small><?php _e('Captures most recent shipping rate lookup.  Recommended to be disabled when website development is complete. This option does not display any messages on frontend.', 'woocommerce-canadapost-webservice') ?></small></label>
						<?php if ($this->options->log_enable): ?>
						<div><a href="<?php echo wp_nonce_url( admin_url( 'admin-ajax.php?action=cpwebservice_rates_log_display' ), 'cpwebservice_rates_log_display' ); ?>" title="Display Log" class="button cpwebservice-log-display"><?php _e('Display most recent request','woocommerce-canadapost-webservice'); ?></a> <img src="<?php admin_url(); ?>/wp-admin/images/wpspin_light.gif" alt="" class="cpwebservice-log-display-loading" style="display: none;" /></div>
						<div id="cpwebservice_log_display" style="display:none;padding:5px;margin:5px;border:1px solid #ccc;width:70%">
						<p></p>
						</div>
						<?php endif; ?> 
						</td>
						</tr>
					</table>
					
			<?php 
			}

			function process_admin_options() {

				
				// check for security
				if (!isset($_POST['cpwebservice_options_noncename']) || !wp_verify_nonce($_POST['cpwebservice_options_noncename'], plugin_basename(__FILE__))) 
					return;


				if(isset($_POST['woocommerce_cpwebservice_enabled'])) $this->options->enabled = 'yes'; else $this->options->enabled ='no';
				if(isset($_POST['woocommerce_cpwebservice_title'])) $this->options->title = woocommerce_clean($_POST['woocommerce_cpwebservice_title']);
				if(isset($_POST['woocommerce_cpwebservice_account'])) $this->options->account = woocommerce_clean($_POST['woocommerce_cpwebservice_account']);
				if(isset($_POST['woocommerce_cpwebservice_contractid'])) $this->options->contractid = woocommerce_clean($_POST['woocommerce_cpwebservice_contractid']);
				if(isset($_POST['woocommerce_cpwebservice_api_user'])) $this->options->api_user = woocommerce_clean($_POST['woocommerce_cpwebservice_api_user']);
				if(isset($_POST['woocommerce_cpwebservice_api_key'])) $this->options->api_key = woocommerce_clean($_POST['woocommerce_cpwebservice_api_key']);
				if(isset($_POST['woocommerce_cpwebservice_source_postalcode'])) $this->options->source_postalcode = woocommerce_clean($_POST['woocommerce_cpwebservice_source_postalcode']);
				$this->options->source_postalcode = str_replace(' ','',strtoupper($this->options->source_postalcode)); // N0N0N0 format only
				if(isset($_POST['woocommerce_cpwebservice_mode'])) $this->options->mode = woocommerce_clean($_POST['woocommerce_cpwebservice_mode']);
				if(isset($_POST['woocommerce_cpwebservice_delivery'])) $this->options->delivery = intval(woocommerce_clean($_POST['woocommerce_cpwebservice_delivery'])); 
				if ($this->options->delivery==0) { $this->options->delivery = ''; }
				if(isset($_POST['woocommerce_cpwebservice_delivery_guarantee'])) $this->options->delivery_guarantee = true; else $this->options->delivery_guarantee = false;
				if(isset($_POST['woocommerce_cpwebservice_margin_value']))  $this->options->margin_value = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_margin_value']));
				if(isset($_POST['woocommerce_cpwebservice_margin'])) $this->options->margin = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_margin']));
				if (!empty($this->options->margin) && $this->options->margin == 0) { $this->options->margin = ''; } // percentage only != 0
				if(isset($_POST['woocommerce_cpwebservice_packageweight'])) $this->options->packageweight = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_packageweight']));
				if(isset($_POST['woocommerce_cpwebservice_display_errors'])) $this->options->display_errors = true; else $this->options->display_errors = false;
				if(isset($_POST['woocommerce_cpwebservice_log_enable'])) $this->options->log_enable = true; else $this->options->log_enable = false;
				if(isset($_POST['woocommerce_cpwebservice_boxes_enable'])) $this->options->boxes_enable = true; else $this->options->boxes_enable = false;
				if(isset($_POST['woocommerce_cpwebservice_lettermail_enable'])) $this->options->lettermail_enable = true; else $this->options->lettermail_enable = false;
				if(isset($_POST['woocommerce_cpwebservice_shipping_tracking'])) $this->options->shipping_tracking = true; else $this->options->shipping_tracking = false;
				if(isset($_POST['woocommerce_cpwebservice_email_tracking'])) $this->options->email_tracking = true; else $this->options->email_tracking = false;
				if(isset($_POST['woocommerce_cpwebservice_tracking_icons'])) $this->options->tracking_icons = true; else $this->options->tracking_icons = false;
				if(isset($_POST['woocommerce_cpwebservice_display_required_notice'])) $this->options->display_required_notice = true; else $this->options->display_required_notice = false;
				update_option('cpwebservice_require_postal', $this->options->display_required_notice  ? 'yes' : 'no');
				if(isset($_POST['woocommerce_cpwebservice_enable_rules'])) $this->options->rules_enable = true; else $this->options->rules_enable = false;
				if(isset($_POST['woocommerce_cpwebservice_product_shipping_options'])) $this->options->product_shipping_options = true; else $this->options->product_shipping_options = false;
				if(isset($_POST['woocommerce_cpwebservice_volumetric_weight'])) $this->options->volumetric_weight = true; else $this->options->volumetric_weight = false;
				if(isset($_POST['woocommerce_cpwebservice_weight_only_enabled'])) $this->options->weight_only_enabled = true; else $this->options->weight_only_enabled = false;
				if(isset($_POST['woocommerce_cpwebservice_prefer_service'])) $this->options->prefer_service = true; else $this->options->prefer_service = false;
				
				// services
				if(isset($_POST['woocommerce_cpwebservice_services']) && is_array($_POST['woocommerce_cpwebservice_services'])) {
					// save valid options. ( returns an array containing all the values of array1 that are present in array2 - in this case, an array of valid service codes)
					$this->services = array_intersect($_POST['woocommerce_cpwebservice_services'], array_keys($this->available_services));
					update_option('woocommerce_cpwebservice_services', $this->services);
 				}
				
				// boxes
				if( isset($_POST) && isset($_POST['woocommerce_cpwebservice_box_length']) && is_array($_POST['woocommerce_cpwebservice_box_length']) ) {
					$boxes = array();
				
					for ($i=0; $i < count($_POST['woocommerce_cpwebservice_box_length']); $i++){
						$box = array();
						$box['length'] = isset($_POST['woocommerce_cpwebservice_box_length'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_box_length'][$i]),1,'.','') : '';
						$box['width'] = isset($_POST['woocommerce_cpwebservice_box_width'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_box_width'][$i]),1,'.','') : '';
						$box['height'] = isset($_POST['woocommerce_cpwebservice_box_height'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_box_height'][$i]),1,'.','') : '';
						$box['name'] = isset($_POST['woocommerce_cpwebservice_box_name'][$i]) ? woocommerce_clean($_POST['woocommerce_cpwebservice_box_name'][$i]) : '';
						$box['weight'] = isset($_POST['woocommerce_cpwebservice_box_weight'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_box_weight'][$i]),1,'.','') : '';
						$box['margin'] = isset($_POST['woocommerce_cpwebservice_box_margin'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_box_margin'][$i]),1,'.','') : '';
						if (empty($box['weight'])) { $box['weight'] = ''; }
						if (empty($box['margin'])) { $box['margin'] = ''; }
						// Cubed/volumetric
						$box['cubed'] = $box['length'] * $box['width'] * $box['height'];
						
						$boxes[] = $box;
					}
				
					$this->boxes = $boxes;
					update_option('woocommerce_cpwebservice_boxes', $this->boxes);
				}
				
				// rules
				if( isset($_POST) && isset($_POST['cpwebservice_rule_classes']) && is_array($_POST['cpwebservice_rule_classes']) ) {
					$rules = array();
					for ($i=0; $i < count($_POST['cpwebservice_rule_classes']); $i++){
						$rule = array();
						if (isset($_POST['cpwebservice_rule_classes'][$i]) && isset($_POST['cpwebservice_rule_services'][$i])){
							$rule['shipping_class'] =  woocommerce_clean($_POST['cpwebservice_rule_classes'][$i]);
							$rule['services'] = array();
							foreach($_POST['cpwebservice_rule_services'][$i] as $svc){
								$rule['services'][] = woocommerce_clean($svc);
							}
						}
						$rules[] = $rule;
					}
				
					$this->rules = $rules;
					update_option('woocommerce_cpwebservice_rules', $this->rules);
				}
				
				// lettermail
				if( isset($_POST) && isset($_POST['woocommerce_cpwebservice_lettermail_country']) && is_array($_POST['woocommerce_cpwebservice_lettermail_country']) ) {
					$lettermail = array();
				
					for ($i=0; $i < count($_POST['woocommerce_cpwebservice_lettermail_country']); $i++){
						$row = array();
						$row['country'] = isset($_POST['woocommerce_cpwebservice_lettermail_country'][$i]) ? woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_country'][$i]) : '';
						$row['label'] = isset($_POST['woocommerce_cpwebservice_lettermail_label'][$i]) ? woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_label'][$i]) : '';
						$row['cost'] = isset($_POST['woocommerce_cpwebservice_lettermail_cost'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_lettermail_cost'][$i]),2,'.','') : '';
						$row['weight_from'] = isset($_POST['woocommerce_cpwebservice_lettermail_weight_from'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_lettermail_weight_from'][$i]),2,'.','') : '';
						$row['weight_to'] = isset($_POST['woocommerce_cpwebservice_lettermail_weight_to'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_lettermail_weight_to'][$i]),2,'.','') : '';
						if ($row['weight_from'] > $row['weight_to']) { $row['weight_from'] = $row['weight_to']; } // Weight From must be a lesser value.
						$row['max_qty'] = isset($_POST['woocommerce_cpwebservice_lettermail_max_qty'][$i]) ? intval($_POST['woocommerce_cpwebservice_lettermail_max_qty'][$i]) : 0;
						$row['min_total'] = !empty($_POST['woocommerce_cpwebservice_lettermail_min_total'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_lettermail_min_total'][$i]),2,'.','') : '';
						$row['max_total'] = !empty($_POST['woocommerce_cpwebservice_lettermail_max_total'][$i]) ? number_format(floatval($_POST['woocommerce_cpwebservice_lettermail_max_total'][$i]),2,'.','') : '';
						
						$lettermail[] = $row;
					}
				
					$this->lettermail = $lettermail;
					update_option('woocommerce_cpwebservice_lettermail', $this->lettermail);
				}
				if(isset($_POST['woocommerce_cpwebservice_lettermail_limits'])) $this->options->lettermail_limits = true; else $this->options->lettermail_limits = false;
				if(isset($_POST['woocommerce_cpwebservice_lettermail_maxlength'])) $this->options->lettermail_maxlength = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_maxlength']));
				if(isset($_POST['woocommerce_cpwebservice_lettermail_maxwidth'])) $this->options->lettermail_maxwidth = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_maxwidth']));
				if(isset($_POST['woocommerce_cpwebservice_lettermail_maxheight'])) $this->options->lettermail_maxheight = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_maxheight']));
				if (empty($this->options->lettermail_maxlength)) $this->options->lettermail_maxlength = '';
				if (empty($this->options->lettermail_maxwidth)) $this->options->lettermail_maxwidth = '';
				if (empty($this->options->lettermail_maxheight)) $this->options->lettermail_maxheight = '';
				if(isset($_POST['woocommerce_cpwebservice_lettermail_packageweight'])) $this->options->lettermail_packageweight = floatval(woocommerce_clean($_POST['woocommerce_cpwebservice_lettermail_packageweight']));
				if(isset($_POST['woocommerce_cpwebservice_lettermail_override_weight'])) $this->options->lettermail_override_weight = true; else $this->options->lettermail_override_weight = false;
				
				// update options.
				update_option('woocommerce_cpwebservice', $this->options);
			}
			
			
			/**
			 * Ajax function to Display Rates Lookup Log.
			 */
			public function rates_log_display() {
			
				// Let the backend only access the page
				if( !is_admin() ) {
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				}
					
				// Check the user privileges
				if( !current_user_can( 'manage_woocommerce_orders' ) && !current_user_can( 'edit_shop_orders' ) ) {
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				}
					
				// Nonce.
				if ( ! isset( $_GET['_wpnonce'] ) || ! wp_verify_nonce( $_GET['_wpnonce'], 'cpwebservice_rates_log_display' ) )
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				
				
				if (false !== ( $log = get_transient('cpwebservice_log') ) && !empty($log)){
				$log = (object) array_merge(array('cart'=>array(),'params'=>array(),'request'=>array(),'rates'=>array(), 'datestamp'=>''), (array) $log);
					?>
										<h4><?php _e('Cart Shipping Rates Request', 'woocommerce-canadapost-webservice')?> - <?php echo date("F j, Y, g:i a",$log->datestamp); ?></h4>
										<table class="table widefat">
										<tr><th><?php _e('Item', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Qty', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Weight', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Dimensions', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Cubic', 'woocommerce-canadapost-webservice')?></th></tr>
										<?php foreach($log->cart as $cart):?>
										<tr>
										<td><?php echo edit_post_link(esc_html($cart['item']),'','',$cart['id'])?></td><td><?php echo esc_html($cart['quantity'])?></td><td><?php echo esc_html($cart['weight'])?>kg</td><td><?php echo esc_html($cart['length'])?>cm x <?php echo esc_html($cart['width'])?>cm x <?php echo esc_html($cart['height'])?>cm</td><td><?php echo esc_html($cart['cubic'])?>cm<sup>3</sup></td>
										</tr>
										<?php endforeach; //$this->log->cart[] = array('quantity'=>$values['quantity'], 'weight'=>$this->convertWeight($values['data']->get_weight(), $weight_unit) * $values['quantity'], 'length'=>$length, 'width'=>$width, 'height'=>$height, 'cubic'=>$cubic); ?>
										</table>
										
										<h4><?php _e('Request / API Response', 'woocommerce-canadapost-webservice')?></h4>
										<p class="description"><?php _e('After box packing/Volumetric weight calculation and Box/Envelope Weight', 'woocommerce-canadapost-webservice')?></p>
										<table class="table widefat">
										<tr><th><?php _e('Packages', 'woocommerce-canadapost-webservice')?> (<?php echo count($log->params)?>)</th><th><?php _e('Country', 'woocommerce-canadapost-webservice')?>, <?php _e('State', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Destination', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Shipping Weight', 'woocommerce-canadapost-webservice')?></th><th><?php _e('Dimensions', 'woocommerce-canadapost-webservice')?></th></tr>
										<?php foreach($log->params as $i=>$params) { ?>
										<tr>
										<td><?php echo (($i+1) . (isset($params['box_name']) ? esc_html(': '.$params['box_name']) : '')) ?></td><td><?php echo esc_html($params['country'])?>, <?php echo esc_html($params['state'])?></td><td><?php echo esc_html($params['postal'])?></td><td><?php echo esc_html(number_format($params['shipping_weight'],2))?>kg (<?php _e('Actual', 'woocommerce-canadapost-webservice')?> <?php echo esc_html(number_format($params['total_weight'],2))?>kg)</td>
										<td><?php echo esc_html(number_format(floatval($params['length']),2))?>cm x <?php echo esc_html(number_format(floatval($params['width']),2))?>cm x <?php echo esc_html(number_format(floatval($params['height']),2))?>cm</td>
										<?php //array('country'=>$country, 'state'=>$state, 'postal'=>$postal, 'shipping_weight'=>$shipping_weight, 'length'=>$length, 'width'=>$width, 'height'=>$height); ?>
										</tr>
										<?php } //endforeach ?>
										</table>
										<br />
										<table class="table widefat">
										<tr><td>
										<?php foreach($log->request as $request):?><?php echo str_replace("\n\n","</td><td>",str_replace("\n","<br />",esc_html($request))) ?><?php endforeach; ?>
										</td>
										</tr></table>
										
										<h4><?php _e('Rates displayed in Cart', 'woocommerce-canadapost-webservice')?></h4>
										<?php if(!empty($log->rates)): ?>
										<table class="table widefat">
										<?php foreach($log->rates as $rates):?>
										<tr>
										<th><?php echo $rates->label ?></th>
										<td><?php echo number_format((float)$rates->cost, 2) ?>
										</td>
										</tr>
										<?php endforeach; ?>
										</table>
										<?php else: ?>
										<p><?php _e('No rates displayed', 'woocommerce-canadapost-webservice') ?></p>
										<?php endif; ?>
										<?php } else { ?>
						<?php _e('No log information.. yet.  Go to your shopping cart page and click on "Calculate Shipping".', 'woocommerce-purolator-webservice') ?>
						<?php  } // endif
				
			exit;
			}
			
			public function shipping_class_rule(){
				if (empty($this->rules) || !is_array($this->rules)){
					$this->rules = array();
				}
				// ensure there are 3 rules available (can change to any number as it's unlimited)
				$display_rules_num = 3;
				if (count($this->rules) < $display_rules_num){
					for($i=count($this->rules);$i<$display_rules_num;$i++){
						$this->rules[] = array('shipping_class'=>'', 'services'=>array(''));
					}
				}
				$shipping_class = get_terms(array('product_shipping_class'), array('hide_empty' => 0));
				?><table class="form-table" style="width:auto">
				<?php foreach($this->rules as $i => $rule){ ?>
				<tr class="cpwebservice_rules">
				<td>
				<select name="cpwebservice_rule_classes[<?php echo $i ?>]" data-placeholder="<?php _e('Choose a Shipping Class...' , 'woocommerce-canadapost-webservice' )?>" class="chosen_select">
            	<option value=""></option>
            <?php foreach($shipping_class as $ship) { ?>
					<option value="<?php echo esc_attr($ship->term_id)?>" <?php selected(isset($rule['shipping_class']) && $rule['shipping_class']==$ship->term_id)?>><?php echo esc_html($ship->name)?></option>
			<?php 	}//end foreach ?>
            </select>
            </td><td>
            	<?php $current_group = ''; ?>
				<?php _e('Can only use', 'woocommerce-canadapost-webservice')?>:
				 <select name="cpwebservice_rule_services[<?php echo $i ?>][]" data-placeholder="<?php _e('Choose Canada Post Services...' , 'woocommerce-canadapost-webservice' )?>" class="widefat chosen_select" multiple>
		            <option value=""></option>
		            <?php foreach($this->available_services as $service_code => $label) { ?>
		            <?php 
		            	$group = $this->get_destination_from_service($service_code);
		            	if ($current_group != $group) { ?>
		            	<?php if ($current_group != '') { ?></optgroup><?php } // endif; ?>
		            	<?php $current_group = $group; ?>
		            	<optgroup label="<?php echo esc_attr($group)?>">
		            	<?php } // endif ?>
		              		<option value="<?php echo esc_attr($service_code)?>" <?php selected(isset($rule['services']) && is_array($rule['services']) && in_array($service_code,$rule['services']))?>><?php echo esc_html($label)?></option>
		            <?php } // endforeach ?>
		            <?php if ($current_group != '') { ?>
		            	</optgroup>
		            	<?php } // endif; ?>
		          </select>
		          </td><td>
		          <a href="javascript:;" class="button-secondary btn_cpwebservice_rules"><?php _e('Clear','woocommerce-canadapost-webservice'); ?></a>
		          </td> 
		          </tr>
				<?php
				}// end foreach  ?>
				</table>
				<?php 
			}
			
			public function get_limited_services($products){
				$services = array();
				$services_rules = array();
				if (isset($this->rules) && is_array($this->rules) && count($this->rules) > 0){
					$ids = $this->get_product_ids($products);
					if (!empty($ids)){
						$term_list = wp_get_object_terms($ids, 'product_shipping_class', array('fields' => 'ids'));
						// Check to see if these products are in a Shipping Class.
						if (!empty($term_list) && is_array($term_list)){
							// In Shipping Class. Check rules table.
							foreach($this->rules as $rule){
								if (isset($rule['shipping_class']) && isset($rule['services']) && is_array($rule['services'])){
									if (in_array($rule['shipping_class'], $term_list)){
										// Shipping Class matches a rule.
										$services_rules[] = $rule['services'];
									}
								}
							}				
						}
					}
				}
				if (count($services_rules) > 1){
					//Only use values that are present in _all_ rules.  array_intersect all rule arrays.
					$services=call_user_func_array('array_intersect', $services_rules);
				} elseif (count($services_rules) == 1){
					$services= $services_rules[0];
				}
				return $services;
			}
			
			// Get item_ids into an easy-to-use array.
			private function get_product_ids($products){
				$ids = array();
				foreach($products as $level){
					foreach($level as $p){
						if (isset($p['item_id'])){
							$ids[] = $p['item_id'];
						}
					}
				}
				$ids = array_unique($ids);
				return $ids;
			}
			
			// Converts a product array to a container array.
			private function to_container($product){
				// put product into its own ['products']
				$container = $product;
				// products are inside a level array.
				$container['products'] = array(array($product));
				return $container;
			}
			
			/**
			 * Load and generate the template output with ajax
			 */
			public function validate_api_credentials() {
				// Let the backend only access the page
				if( !is_admin() ) {
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				}
					
				// Check the user privileges
				if( !current_user_can( 'manage_woocommerce_orders' ) && !current_user_can( 'edit_shop_orders' ) ) {
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				}
					
				// Check the nonce
				if( empty( $_GET['action'] ) || !check_admin_referer( $_GET['action'] ) ) {
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
				}
					
				// Check if all parameters are set
				if( empty( $_POST['api_user'] ) || empty( $_POST['api_key'] ) || empty( $_POST['customerid'] ) ) {
					wp_die( __( 'API username, API password, Customer ID and Origin Postal Code are required.' , 'woocommerce-purolator-webservice' ) );
				}
			
				// Nonce.
				if ( ! isset( $_GET['_wpnonce'] ) || ! wp_verify_nonce( $_GET['_wpnonce'], 'cpwebservice_validate_api_credentials' ) )
					wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
					
				// Get api_user, api_key, customerid
				$api_user = sanitize_text_field( $_POST['api_user'] );
				$api_key = sanitize_text_field( $_POST['api_key'] );
				$customerid = sanitize_text_field( $_POST['customerid'] );
				$contractid = sanitize_text_field( $_POST['contractid'] );
			
				$apiValid = false;
				$message = "";
				
				// Check API.
				$username = $this->options->api_user = $api_user;
				$password = $this->options->api_key = $api_key;
				$this->options->account = $customerid;
				$this->options->contractid = $contractid;
				
				
				// REST URL  (Get Service Info)
				$service_url = ($this->options->mode=='live') ? 'https://soa-gw.canadapost.ca/rs/ship/service/DOM.EP?country=CA' : 'https://ct.soa-gw.canadapost.ca/rs/ship/service/DOM.EP?country=CA';
				
				// If using WPML:
				if (defined('ICL_LANGUAGE_CODE')){
					$service_language = (ICL_LANGUAGE_CODE=='fr') ? 'fr-CA':'en-CA'; // 'en-CA' is default
				} else if (get_locale() == 'fr_FR'){ 
					$service_language = 'fr-CA';
				} else {
					$service_language = 'en-CA';
				}
				
				echo ($this->options->mode=='live') ? 'Production/Live Server: ' : 'Development Server: ';
				
				if (!empty($username) && !empty($password)){ 
					try {
						$curl = curl_init($service_url); // Create REST Request
						curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);
						curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, 2);
						curl_setopt($curl, CURLOPT_CAINFO, CPWEBSERVICE_PLUGIN_PATH . '/cert/cacert.pem'); // Signer Certificate in PEM format
						curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
						curl_setopt($curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
						curl_setopt($curl, CURLOPT_USERPWD, $username . ':' . $password);
						curl_setopt($curl, CURLOPT_HTTPHEADER, array('Accept: application/vnd.cpc.ship.rate-v2+xml','Accept-language: '.$service_language));
						$curl_response = curl_exec($curl); // Execute REST Request
						if(curl_errno($curl)){
							$message .= 'Error: ' . curl_error($curl) . "\n";
						}
						$message .= 'HTTP Response Status: ' . curl_getinfo($curl,CURLINFO_HTTP_CODE) . "\n";
						curl_close($curl);
						
						// Using SimpleXML to parse xml response
						libxml_use_internal_errors(true);
						$xml = simplexml_load_string('<root>' . preg_replace('/<\?xml.*\?>/','',$curl_response) . '</root>');
						if (!$xml) {
							$errmsg = 'Failed loading XML' . "\n";
							$errmsg .= $curl_response . "\n";
							foreach(libxml_get_errors() as $error) {
								$errmsg.= "\t" . $error->message;
							}
							$message .= $errmsg;
						} else {
							if ($xml->{'service'} ) {
								// Success! API correctly responded.
								$apiValid = true;
							}
							if ($xml->{'messages'}) {
								$apierror = '';
								$messages = $xml->{'messages'}->children('http://www.canadapost.ca/ws/messages');
								foreach ( $messages as $message ) {
									$apierror .= 'Error Code: ' . $message->code . "\n";
									$apierror .= 'Error Msg: ' . $message->description . "\n\n";
								}
								$message .= $apierror;
							}
						}
						
					} catch (Exception $ex) {
						// cURL request went wrong.
						$message .= 'Error: ' . $ex . "\n";
					}
				}
				
				echo str_replace("\n","<br />",$message);
				
				if ($apiValid) {
					echo '<strong>Success!</strong> API Credentials validated with Canada Post.';
				} else {
					echo '<strong>Failed</strong> API Credentials did not validate.';
				}
				
				// Try get_rates to see if customer info works.
				if ($apiValid) {
					echo '<br /><strong>Testing Rates Lookup:</strong><br />';
					$this->options->display_errors = true;
					$rates = $this->get_rates('CA','ON','K1A 0B1',0.5,5,5,2); // Ship 5x5x2cm package to CP headquarters, Ottawa.
					if (is_array($rates) && !empty($rates)) {
						echo '<br /><strong>Rates Lookup Success!</strong> CustomerID/Venture One information appears to be valid.';
					} else {
						echo '<br /><strong>Rates Lookup Failed</strong> Unable to look up rates. CustomerID/Venture One account number may be invalid or inactive.';
					}
				}
				
				exit;
			}
			
			
			function boxFits($box, $max){
				if (is_array($box) && is_array($max)){
					$fitbox = array($box['width'], $box['height'], $box['length']);
					$fitmax = array($max['width'], $max['height'], $max['length']);
					
					$it_fits = true;
					
					while (count($fitbox) > 0){
						// Compare the maximum dimensions with each other.
						// If the MaxDim is bigger than the Box Dim, then it doesn't fit.
						$b = max($fitbox);
						$m = max($fitmax);
						if ($m > $b){
							$it_fits = false;
							break;
						} else {
							// Remove value from each array.
							// Remove one max value from Box
							$b_index = array_search($b, $fitbox);
							if (false !== $b_index) {
								unset($fitbox[$b_index]);
							} else { $it_fits = false; break; } // safety so loop doesn't continue.
							// Remove one max value from MaxDimension
							$m_index = array_search($m, $fitmax);
							if (false !== $m_index) {
								unset($fitmax[$m_index]);
							} else { $it_fits = false; break; } // safety	
						}
					}
					
					return $it_fits;
				}		
				return false;
			}
			
			function convertSize($size,$unit_from) {
				$size = floatval($size);
				$finalSize = $size;
				switch ($unit_from) {
					// we need the units in cm
					case 'cm':
						// change nothing
						$finalSize = $size;
						break;
					case 'in':
						// convert from in to cm
						$finalSize = $size * 2.54;
						break;
					case 'yd':
						// convert from yd to cm
						$finalSize = $size * 3 * 2.54;
						break;
					case 'm':
						// convert from m to cm
						$finalSize = $size * 100;
						break;
					case 'mm':
						// convert from mm to cm
						$finalSize = $size * 0.1;
						break;
				}
				return $finalSize;
			}
			
			function convertWeight($weight,$unit_from) {
				$finalWeight = $weight;
				switch ($unit_from) {
					// we need the units in kg
					case 'kg':
						// change nothing
						$finalWeight = $weight;
						break;
					case 'g':
						// convert from g to kg
						$finalWeight = $weight * 0.001;
						break;
					case 'lbs':
						// convert from lbs to kg
						$finalWeight = $weight * 0.4535;
						break;
					case 'oz':
						// convert from oz to kg
						$finalWeight = $weight * 0.0283;
						break;
				}
				return $finalWeight;
			}

}