<?php
/*
 Order Details class
woocommerce_cpwebservice_orderdetails.php

Copyright (c) 2013-2015 Jamez Picard

*/
class woocommerce_cpwebservice_orderdetails
{
	
	public $method_id = 'woocommerce_cpwebservice';
	
	/**
	 * __construct function.
	 *
	 * @access public
	 * @return woocommerce_cpwebservice_orderdetails
	 */
	function __construct() {

		$this->init();
	}
	
	/*
	 * Init
	 */
	function init() {
		$this->options = get_option('woocommerce_cpwebservice', array('enabled'=>'no'));
		if ($this->options->enabled){
			// Wire up actions
			if (is_admin()){
				add_action( 'add_meta_boxes', array(&$this, 'add_shipping_details_box') );
				wp_enqueue_script( 'cpwebservice_modal' ,plugins_url( 'lib/modal.js' , dirname(__FILE__) ) , array( 'jquery' ) );
				wp_enqueue_style( 'cpwebservice_woocommerce_admin' , plugins_url( 'lib/admin.css' , dirname(__FILE__) ) );
			} else {
				add_action( 'cpwebservice_order_shipping_info', array(&$this, 'order_shipping_info'), 10, 1 );
				add_action( 'woocommerce_cart_emptied', array(&$this, 'order_shipping_info_reset'));
				// The following 2 are not used as these are wired up earlier.
				//add_action( 'woocommerce_order_add_shipping', array(&$this, 'order_add_shipping'), 10, 3 );
				//add_action( 'woocommerce_checkout_order_processed', array(&$this, 'order_processed') , 10, 2);
			}
		}
	}
	
	/* 
	 * Shipping rate data saved in session
	 */
	function order_shipping_info($shipping_info = array()){
		// Keep $packages in session.
		WC()->session->set( 'cpwebservice_shipping_info', $shipping_info );
	}
	
	/*
	 Hooks: //do_action( 'woocommerce_order_add_shipping', $this->id, $item_id, $shipping_rate );
	 */
	function order_add_shipping($id, $item_id, $shipping_rate) {
		
		// Only process data if the order is placed with this shipping method.
		if ($shipping_rate->method_id == $this->method_id){
			// Retrieve from session
			$shipping_info = WC()->session->get( 'cpwebservice_shipping_info' );
			if (!empty($shipping_info) && is_array($shipping_info)){
				// Add selected shipping.
				if (isset($shipping_info['rates']) && is_array($shipping_info['rates'])) {
					foreach($shipping_info['rates'] as $rate){
						if ($this->method_id.':'.$rate->service_code == $shipping_rate->id){
							$shipping_rate->guaranteed = $rate->guaranteed;
							$shipping_rate->expected_delivery = $rate->expected_delivery;
							$shipping_rate->expected_mailing_date = $rate->expected_mailing_date;
							$shipping_rate->service_code = $rate->service_code;
						}
					}
				}
				// Replace rates with rate used
				$shipping_info['rates'] = $shipping_rate;
				$shipping_info['item_id'] = $item_id;
				// Save with order.
				update_post_meta($id, '_cpwebservice_shipping_info', $shipping_info);
			}
		}
	}
	
	/*
	 * Save Shipping rate data to Order info.
	* 
	*/
	/* function order_processed($order_id, $orderdata) {
		$shipping_info = WC ()->session->get ( 'cpwebservice_shipping_info' );
		if (! empty ( $shipping_info ) && is_array ( $shipping_info )) {
			// Add selected shipping.
			$shipping_methods = WC ()->session->get ( 'chosen_shipping_methods' );
			// Check to ensure the chosen shipping_method is one managed by this plugin.
			$selected_method = null;
			if (!empty($shipping_methods) && is_array($shipping_methods) && !empty($shipping_info['rates']) && is_array($shipping_info['rates'])) {
				foreach($shipping_info['rates'] as $rate){
					if (in_array($this->method_id.':'.$rate->service_code, $shipping_methods)) {
						//found!
						$selected_method = $rate;
						break;
					}
				}
			}
			
			if (!empty($selected_method)){
				// Replace rates with rate used
				$shipping_info ['rates'] = $selected_method;
				// Package info should already be in $shipping_info.
				// Save with order.
				update_post_meta ( $order_id, '_cpwebservice_shipping_info', $shipping_info );
			}
		}
	} */
	
	/*
	 * Shipping rate data saved in session
	*/
	function order_shipping_info_reset(){
		if (WC()->session->cpwebservice_shipping_info!=null){
			unset( WC()->session->cpwebservice_shipping_info );
		}
	}
	
	public function add_shipping_details_box() {
		global $post_id;
		$shipping_info = get_post_meta( $post_id, '_cpwebservice_shipping_info', true);
		if (!empty($shipping_info)){
			add_meta_box( 'cpwebservice_shipping_details', __( 'Order Shipping Details', 'woocommerce-canadapost-webservice' ),  array(&$this,'display_shipping_view'), 'shop_order', 'normal', 'default' );
		}
	}
	
	public function display_shipping_view(){
		global $post_id;
		?>
		<div><img src="<?php echo plugins_url( 'img/canada-post.jpg' , dirname(__FILE__) ); ?>" /></div>
<div id="cpwebservice_shipping_info">
			<?php 
			// Shipping information. array('rates'=>array(),'packages'=>array())
			$shipping_info = get_post_meta( $post_id, '_cpwebservice_shipping_info', true);
			if (!empty($shipping_info)){
			
			if (isset($shipping_info['rates']) && is_object($shipping_info['rates'])) {
				// Get selected rate.
				?>
								<p><strong><?php _e('Canada Post', 'woocommerce-canadapost-webservice'); ?> <?php echo !empty($shipping_info['rates']->label) ? $shipping_info['rates']->label : '' ?></strong> 
									<?php echo !empty($shipping_info['rates']->expected_delivery) ? '<br />'.( !empty($shipping_info['rates']->guaranteed) ? __('Guaranteed', 'woocommerce-canadapost-webservice') . ' ' : '') .__('Delivered by', 'woocommerce-canadapost-webservice'). ' ' .$shipping_info['rates']->expected_delivery : ''  ?>
									<?php echo (!empty($shipping_info['rates']->expected_delivery) && !empty($shipping_info['rates']->expected_mailing_date)) ? __('if mailed by').': ' . $shipping_info['rates']->expected_mailing_date : '' ?> 
								</p>
								<?php 
			}
			if (isset($shipping_info['packages']) && is_array($shipping_info['packages'])) {
					?><h4> <?php _e('Packages', 'woocommerce-canadapost-webservice') ?> (<?php echo count($shipping_info['packages']) ?>)
					<?php _e('to be shipped by', 'woocommerce-canadapost-webservice') ?> <?php _e('Canada Post', 'woocommerce-canadapost-webservice') ?></h4>
					
	<table class="widefat">
		<tr>
			<th width="10%" style="display: none"></th>
			<th><?php _e('Dimensions', 'woocommerce-canadapost-webservice'); ?></th>
			<th><?php _e('Shipping Weight', 'woocommerce-canadapost-webservice'); ?>, <?php _e('Volume/Cubic', 'woocommerce-canadapost-webservice'); ?></th>
			<th><?php _e('Service', 'woocommerce-canadapost-webservice'); ?></th>
			<th><?php _e('Products Packed', 'woocommerce-canadapost-webservice'); ?></th>
		</tr>
					<?php foreach($shipping_info['packages'] as $packages ){ ?>
						<?php if (isset($packages['length']) && isset($packages['width']) && isset($packages['height']) && isset($packages['weight'])){ ?>
			<tr>
			<td class="cpwebservice_order_actions" style="display: none"><a href="<?php echo wp_nonce_url( admin_url( 'admin-ajax.php?action=cpwebservice_create_shipment&order_id=' . $post_id ), 'cpwebservice_create_shipment' ); ?>" class="button button-primary cpwebservice_iframe_modal" target="_blank">
				<?php _e('Create Shipment', 'woocommerce-canadapost-webservice'); ?></a>
			<td><?php echo esc_html($packages['length'] .' x '. $packages['width'] .' x ' . $packages['height'] . ' cm' )?> 
				<?php echo isset($packages['box_name']) ? ' (' . __('Box Name', 'woocommerce-canadapost-webservice').': '.esc_html($packages['box_name']).')' : ''?>
				<?php echo (isset($packages['prepacked']) && $packages['prepacked']=='yes') ? ' ('.__('Prepackaged', 'woocommerce-canadapost-webservice').')' : ''?> 
			</td>
			<td><?php echo esc_html(number_format(floatval($packages['weight']),2,'.','') . ' kg')?> <?php echo isset($packages['actual_weight']) ? esc_html('('.__('Actual', 'woocommerce-canadapost-webservice') .': '.number_format(floatval($packages['actual_weight']),2,'.','') . ' kg)') : '' ?> 
			<br /> <?php echo isset($packages['cubic']) ? esc_html(number_format(floatval($packages['cubic']),0,'.','')) . ' cm<sup>3</sup>' : ''?></td>
			<td><?php
			$selected_rate = null;
			if (isset($shipping_info['rates']) && is_object($shipping_info['rates'])) {
				$selected_rate = $shipping_info['rates']->service_code;
			}
				if (!empty($selected_rate) && !empty($packages['rate']) && is_array($packages['rate'])) {
						foreach($packages['rate'] as $itemrate){
							if ($itemrate->service_code == $selected_rate){
							?>
							<p> <strong><?php _e('Canada Post', 'woocommerce-canadapost-webservice'); ?> <?php echo !empty($itemrate->service) ? $itemrate->service : '' ?></strong> 
							<br /><?php _e('My Cost: ', 'woocommerce-canadapost-webservice') ?> <?php echo !empty($itemrate->price) ? number_format(floatval($itemrate->price),2,'.','') : '' ?>
								<?php echo !empty($itemrate->expected_delivery) ? '<br />'.( !empty($itemrate->guaranteed) ? __('Guaranteed', 'woocommerce-canadapost-webservice') . ' ' : '') .__('Delivered by', 'woocommerce-canadapost-webservice'). ' ' .$itemrate->expected_delivery : ''  ?>
							</p>
							<?php
							}
						}// endforeach

				} // endif
			?></td>
			<td><?php if (!empty($packages['products']) && is_array($packages['products'])) { ?>
				<?php 
				$product_reference = $this->get_product_array($packages['products']);
				// Display
				$product_groups = $this->group_products($packages['products']);
				foreach($product_groups as $item){ 
					if (isset($item['item_id']) && isset($product_reference[$item['item_id']])){
					$p = $product_reference[$item['item_id']];
					?>
					&bull; <?php echo esc_html($item['count']); ?>x <a href="<?php echo $p->get_permalink(); ?>"><?php echo $p->get_title(); ?></a> <?php echo esc_html($p->get_sku()); ?> (<?php echo esc_html($p->get_dimensions()); ?> &nbsp; <?php echo esc_html($p->get_weight() . ' ' . get_option('woocommerce_weight_unit')); ?>) <br />
					<?php 
					}// endif
				}// end foreach
				?>
				<?php } // endif ?>
			</td>
			</tr>
			<?php } // endif ?>
					<?php } // end foreach ?>
					</table>
					<?php 
				}// end if
				
			
			} else {
				echo '<p>' . __('No information available', 'woocommerce-canadapost-webservice') . '</p>'; 
			}
			?>
		</div>
<?php 
	}
	
	// Get item_ids into an easy-to-use array.
	private function get_product_ids($products){
		$ids = array();
		foreach($products as $level){
			foreach($level as $p){
				if (isset($p['item_id'])){
					$ids[] = $p['item_id'];
				}
			}
		}
		$ids = array_unique($ids);
		return $ids;
	}
	
	// Looks up products in $packages['products']
	private function get_product_array($products){
		// Begin Do Lookup
		$ids = $this->get_product_ids($products);
		$items = get_posts(array('post_type' => array('product','product_variation'), 'post__in' => $ids ));
		$product_reference = array();
		foreach($items as $item){ 
			$p = wc_get_product($item);
			$product_reference[$item->ID] = $p;
		} // endforeach
		wp_reset_postdata();
		// End Do Lookup.
		
		return $product_reference;
	}
	
	// Group (and count) Products by ID
	private function group_products($products){
		$group = array();
		foreach($products as $level){
			foreach($level as $p){
				if (isset($p['item_id'])){
					$id = $p['item_id'];
					if (isset($group[$id])) {
						$group[$id]['count'] += 1;
					} else {
						$group[$id] = $p;
						$group[$id]['count'] = 1;
					}
				}
			}
		}
		return $group;
	}

}