<?php
/*
 Product Options class
woocommerce_cpwebservice_products.php

Copyright (c) 2013-2015 Jamez Picard

*/
class woocommerce_cpwebservice_products
{

	/**
	 * __construct function.
	 *
	 * @access public
	 * @return woocommerce_cpwebservice_products
	 */
	function __construct() {
		$this->init();
	}

	/**
	 * init function.
	 *
	 * @access public
	 * @return void
	 */
	function init() {
		$default_options = (object) array('enabled'=>'no', 'title'=>'Canada Post', 'api_user'=>'', 'api_key'=>'','account'=>'','contractid'=>'','source_postalcode'=>'','mode'=>'live', 'display_errors'=>false, 
						   'delivery'=>'', 'delivery_guarantee'=>false, 'margin'=>'', 'margin_value'=>'', 'packageweight'=>floatval('0.02'), 'boxes_enable'=> false, 'boxes_switch'=>true, 'lettermail_enable'=> false, 'rules_enable'=>false, 'volumetric_weight'=>true, 'product_shipping_options' => true,
						   'shipping_tracking'=> true, 'email_tracking'=> true, 'log_enable'=>false,'lettermail_limits'=>false,'lettermail_maxlength'=>'','lettermail_maxwidth'=>'','lettermail_maxheight'=>'','lettermail_override_weight'=>false,'lettermail_packageweight'=>'', 'tracking_icons'=> true, 'display_required_notice'=>true);
		$this->options		= get_option('woocommerce_cpwebservice', $default_options);
		$this->options		= (object) array_merge((array) $default_options, (array) $this->options); // ensure all keys exist, as defined in default_options.
		$this->enabled		= $this->options->product_shipping_options && ( !empty($this->options->api_user) && !empty($this->options->api_key) );

		if ($this->enabled && $this->options->product_shipping_options) {
			// Actions
			add_action('woocommerce_product_options_shipping',  array(&$this, 'add_product_shipping_box') );
			add_action( 'save_post', array(&$this, 'save_product_shipping'), 10, 2 );
			add_action( 'woocommerce_product_after_variable_attributes',  array(&$this, 'add_product_shipping_box_variation'), 10, 3 );
			add_action( 'woocommerce_save_product_variation',  array(&$this, 'save_product_variation'), 10, 2 );
		}

	}

	

	/* Adds a box to the product Edit screen in the Shipping Tab. */
	public function add_product_shipping_box() {
		global $post;
		$product_shipping = get_post_meta( $post->ID, '_cpwebservice_product_shipping', true );
		?>
		<p class="form-field cpwebservice_product_shipping_field"><label for="cpwebservice_product_shipping"><?php _e( 'Canada Post: Ready to Ship (Pre-packaged):', 'woocommerce' ); ?></label> 
		<input name="cpwebservice_product_shipping" type="checkbox" value="1" <?php checked($product_shipping, 'yes'); ?> /> 
		<img class="help_tip" data-tip="<?php esc_attr_e( 'If this is selected, this item will not be placed in any box but assumed already packed and ready to ship when calculating rates with Canada Post.', 'woocommerce' ); ?>" src="<?php echo esc_url( WC()->plugin_url() ); ?>/assets/images/help.png" height="16" width="16" /></p>
		<?php
	}
	
	// do_action ('save_post')
	public function save_product_shipping($post_id, $post){
		// If this isn't a 'product', don't update it.
		if ( 'product' != $post->post_type ) {
			return;
		}
		// Let the backend only access the page
		if( !is_admin() ) {
			wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
		}
		// Check the user privileges
		if( !current_user_can( 'manage_woocommerce' ) ) {
			return;
		}
		
		$product_shipping     = isset($_POST['cpwebservice_product_shipping'] ) ? 'yes' : 'no';
		// Update post meta
		update_post_meta( $post_id, '_cpwebservice_product_shipping', wc_clean( $product_shipping ) );
	}
	
	// do_action('woocommerce_product_after_variable_attributes',$loop, $variation_data, $variation);
	public function add_product_shipping_box_variation($loop, $variation_data, $variation) {
		$product_shipping = get_post_meta( $variation->ID, '_cpwebservice_product_shipping', true );
		?>
		<tr>
			<td class="hide_if_variation_virtual">
			<label><?php _e( 'Canada Post', 'woocommerce-canadapost-webservice' ); ?>: <?php _e('Ready to Ship (Pre-packaged)', 'woocommerce-canadapost-webservice' )?> <input name="variable_cpwebservice_product_shipping[<?php echo esc_attr($loop) ?>]" type="checkbox" class="checkbox" value="1" <?php checked($product_shipping, 'yes'); ?> /> <img class="help_tip" data-tip="<?php esc_attr_e( 'If this is selected, this item will not be placed in any box but assumed already packed and ready to ship when calculating rates with Canada Post.', 'woocommerce' ); ?>" src="<?php echo esc_url( WC()->plugin_url() ); ?>/assets/images/help.png" height="16" width="16" /></label>
			</td>
			<td class="hide_if_variation_virtual">
				
			</td>
		</tr>
		<?php 
	}

	//do_action( 'woocommerce_save_product_variation', $variation_id, $i );
	// called within a for($i)
	public function save_product_variation($variation_id, $i) {
		$variable_cpwebservice_product_shipping = isset( $_POST['variable_cpwebservice_product_shipping'] ) ? $_POST['variable_cpwebservice_product_shipping'] : array();
		$product_shipping     = isset( $variable_cpwebservice_product_shipping[ $i ] ) ? 'yes' : 'no';
		
		// Update variation
		update_post_meta( $variation_id, '_cpwebservice_product_shipping', wc_clean( $product_shipping ) );
	}
	
	
	
}