<?php
/**
 * Custom Update class
 * Allows plugins to use their own update API.
*
* Based on code by Pippin
*/
class woocommerce_cpwebservice_update {
	private $api_url = 'https://truemedia.ca/plugins/cpwebservice/version/';
	private $api_data = '';
	private $slug = '';
	private $plugin = '';
	private $version = '1.0';
	private $upgrade_url = "http://codecanyon.net/item/canada-post-woocommerce-shipping-plugin/5216356";

	/**
	 * Class constructor.
	 *
	 *
	 * @param string $_api_url The URL pointing to the custom API endpoint.
	 * @param string $_plugin_file Path to the plugin file.
	 * @param array $_api_data Optional data to send with API calls.
	 * @return void
	 */
	function __construct( $_api_url, $_slug, $_api_data = null ) {
		$this->api_url =  $_api_url;
		$this->slug = $_slug; // basename( $_slug, '.php');
		$this->plugin = plugin_basename( $_slug );
		$this->api_data = $_api_data;
		$this->version = $_api_data['version'];
		
		// debug
		//set_site_transient( 'update_plugins', null );
		
		// Set up filters
		//add_filter( 'pre_set_site_transient_update_plugins', array( $this, 'pre_set_site_transient_update_plugins_filter' ) );
		add_filter( 'plugins_api', array( $this, 'plugins_api_filter' ), 10, 3);
		add_action('admin_init', array($this, 'plugin_custom_update'), 30);
	}
	
	/*
	Replace main Upgrade link with custom link.
	 */
	function plugin_custom_update(){
		$plugin_file = $this->plugin;
		if (has_action("after_plugin_row_$plugin_file")){
			remove_action( "after_plugin_row_$plugin_file", 'wp_plugin_update_row', 10, 2 );
			add_action( "after_plugin_row_$plugin_file", array($this, 'wp_plugin_update_row'), 10, 2 );
		}
	}
	/*
	 * Displays update row. (custom wp_plugin_update_row function)
	 */
	function wp_plugin_update_row( $file, $plugin_data ) {
		$current = get_site_transient( 'update_plugins' );
		if ( !isset( $current->response[ $file ] ) )
			return false;
		$r = $current->response[ $file ];
		$plugins_allowedtags = array('a' => array('href' => array(),'title' => array()),'abbr' => array('title' => array()),'acronym' => array('title' => array()),'code' => array(),'em' => array(),'strong' => array());
		$plugin_name = wp_kses( $plugin_data['Name'], $plugins_allowedtags );
		$details_url = $this->upgrade_url;
		$wp_list_table = _get_list_table('WP_Plugins_List_Table');
		if ( is_network_admin() || !is_multisite() ) {
			echo '<tr class="plugin-update-tr"><td colspan="' . $wp_list_table->get_column_count() . '" class="plugin-update colspanchange"><div class="update-message">';
			if ( ! current_user_can('update_plugins') )
				printf( __('There is a new version of %1$s available. <a href="%2$s" title="%3$s">View version %4$s details</a>.'), $plugin_name, esc_url($details_url), esc_attr($plugin_name), $r->new_version );
			else if ( empty($r->package) )
				printf( __('There is a new version of %1$s available. <a href="%2$s" title="%3$s">View version %4$s details</a>. <em>Automatic update is unavailable for this plugin.</em>'), $plugin_name, esc_url($details_url), esc_attr($plugin_name), $r->new_version );
			else
				printf( __('There is a new version of %1$s available. <a href="%2$s" title="%3$s">View version %4$s details</a> or <a href="%5$s">update now</a>.'), $plugin_name, esc_url($details_url), esc_attr($plugin_name), $r->new_version, wp_nonce_url( self_admin_url('update.php?action=upgrade-plugin&plugin=') . $file, 'upgrade-plugin_' . $file) );

			do_action( "in_plugin_update_message-{$file}", $plugin_data, $r );
			echo '</div></td></tr>';
		}
	}
	
	
	/**
	 * Check for Updates at the defined API endpoint and modify the update array.
	 *
	 * This function dives into the update api just when Wordpress creates its update array,
	 * then adds a custom API call and injects the custom plugin data retrieved from the API.
	 * It is reassembled from parts of the native Wordpress plugin update code.
	 * See wp-includes/update.php line 121 for the original wp_update_plugins() function.
	 *
	 * @uses api_request()
	 *
	 * @param array $_transient_data Update array build by Wordpress.
	 * @return array Modified update array with custom plugin data.
	 */
	function pre_set_site_transient_update_plugins_filter( $_transient_data ) {

		if( empty( $_transient_data ) ) return $_transient_data;

		// Add cpwebservice plugin data to update array (if version is less)
		$api_response = $this->api_request( null );

		if( false !== $api_response && is_object( $api_response ) ) {
			
			// update this plugin version
			$this->update_version();
			if( version_compare( $this->version, $api_response->new_version, '<' ) ){				
				$_transient_data->response[$this->slug] = $api_response;
			} else {
				if (isset($_transient_data->response[$this->slug])) 
					unset($_transient_data->response[$this->slug]);
			}
		}
		return $_transient_data;
	}


	/**
	 * Updates information on the "View version x.x details" page with custom data.
	 *
	 * @uses api_request()
	 *
	 * @param mixed $_data
	 * @param string $_action
	 * @param object $_args
	 * @return object $_data
	 */
	function plugins_api_filter( $_data, $_action = '', $_args = null ) {
		if ( ( $_action != 'plugin_information' ) || !isset( $_args->slug ) || ( $_args->slug != $this->slug ) ) return $_data;

		$api_data = array( 'slug' => $this->slug );

		$api_response = $this->api_request( 'plugin_information', $api_data );
		if ( false !== $api_response ) $_data = $api_response;

		return $_data;
	}

	/**
	 * Calls the API and, if successfull, returns the object delivered by the API.
	 *
	 * @uses wp_remote_post()
	 * @uses is_wp_error()
	 *
	 * @param string $_action The requested action.
	 * @param array $_data Parameters for the API action.
	 * @return false||object
	 */
	private function api_request( $_data ) {
		
		//$data = array_merge( $this->api_data, $_data );
		
		$request = wp_remote_post( $this->api_url, array( 'timeout' => 15, 'ssverify' => true, 'body' => '' ) );
		if ( !is_wp_error( $request ) ):
		$request = json_decode( wp_remote_retrieve_body( $request ) );
		if( $request )
			$request->sections = maybe_unserialize( $request->sections );
		return $request;
		else:
		return false;
		endif;
	}
	
	/* gets version from plugin file*/
	private function update_version() {
		$info = get_plugin_data(plugin_dir_path($this->plugin).$this->plugin);
		$this->version = $info['Version'];
	}
}